#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from collections import namedtuple
import tensorflow as tf

from tensorflow.core.framework import graph_pb2
from tensorflow.python.framework import dtypes
from neural_compressor.adaptor.tf_utils.quantize_graph_common import QuantizeGraphHelper as helper
from neural_compressor.adaptor.tf_utils.util import version1_gt_version2
from neural_compressor.adaptor.tf_utils.util import version1_lt_version2
from neural_compressor.adaptor.tf_utils.util import version1_eq_version2

class QuantizeGraphBase():
    """
    This is the base class for quantize graph.
    """

    def __init__(self, output_node_names):
        self.output_node_names = output_node_names
        self.transformers = {}

    def register_transformer(self, node_name, entry):
        if node_name not in self.transformers:
            self.transformers[node_name] = {}

        self.transformers[node_name] = entry

    def do_transform(self):
        """
        This is the virtual interface need to be implemented by derived class
        :return:
        """
        pass

    def remove_dead_nodes(self, input_graph, output_names):
        """Removes nodes that are no longer needed for inference from the graph."""
        return tf.compat.v1.graph_util.extract_sub_graph(
            input_graph, output_names)

class QuantizeNodeBase():
    """This is the base class for nodes fusion


    Arguments:
        object {[type]} -- [description]
    """
    node_details = namedtuple('node_details', ['node', 'output'])

    def __init__(self, **kwargs):

        self.logger = logging.getLogger()
        input_graph = kwargs['input_graph']

        assert isinstance(input_graph, graph_pb2.GraphDef)

        self.input_graph = input_graph

        self._parse_graph()
        self.output_node_maps = {}
        self.output_graph = graph_pb2.GraphDef()
        self.quantized_node_dict = {}
        self.patterns =  kwargs['patterns']
        self.remove_redundant_quant_flag = kwargs['remove_redundant_quant_flag']
        self.fake_quant = kwargs['fake_quant'] if 'fake_quant' in kwargs else False
        self.per_channel, self.is_asymmetric = kwargs['op_wise_cfg'][0], kwargs['op_wise_cfg'][2]
        self.op_wise_config_name_list = kwargs['op_wise_config_name_list']
        self.weight_bit = kwargs['op_wise_cfg'][3]
        self.start_node_name = kwargs['start_node_name']
        self.device = kwargs['device']
        self.new_api = kwargs['new_api']
        self.enable_s8 = bool(version1_gt_version2(tf.version.VERSION, '2.1.0') or \
                    version1_eq_version2(tf.version.VERSION, '2.1.0') or \
            tf.version.VERSION.find('1.15.0-up') != -1)

    def apply_the_transform(self):
        """
        This is the virtual interface to be implemented by derived class
        :return: transformed graphdef
        """
        pass

    def get_longest_fuse(self):
        """This is the virtual interface to be implemented by derived class
        """
        pass

    def _insert_dummy_biasadd(self, match_node_name, matched_node): # pragma: no cover
         target_node_name = matched_node.node.name
         matmul_a_node_name = helper.node_name_from_input(matched_node.node.input[0])
         matmul_a_node = self.node_name_mapping[matmul_a_node_name].node
         matmul_b_node_name = helper.node_name_from_input(matched_node.node.input[1])
         matmul_b_node = self.node_name_mapping[matmul_b_node_name].node

         if matmul_a_node.op == 'Const' and matmul_b_node.op != 'Const':
             pass
         else:
             from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
             g = GraphAnalyzer()
             g.graph = self.input_graph
             graph_info = g.parse_graph()
             next_node_names = graph_info[matched_node.node.name].outputs
             bias_node_name = target_node_name + '_dummy_biasadd'
             bias_const_node_name = target_node_name + '_fake_const'

             if matched_node.node.op == 'MatMul':
                 t_b_index = 0 if matched_node.node.attr['transpose_b'].b else 1
             if matched_node.node.op in ('Conv2D' or 'DepthwiseConv2dNative') and \
                matched_node.node.attr['data_format'].s == b'NHWC':
                 t_b_index = 3
             elif matched_node.node.op in ('Conv2D' or 'DepthwiseConv2dNative')  and \
                matched_node.node.op.attr['data_format'].s == b'NCHW':
                 t_b_index = 1
             elif matched_node.node.op == 'Conv3D' and matched_node.node.attr['data_format'].s == b'NDHWC':
                 t_b_index = 4
             elif matched_node.node.op == 'Conv3D' and matched_node.node.attr['data_format'].s == b'NCDHW':
                 t_b_index = 1
             bias_add_length = matmul_b_node.attr['value'].tensor.tensor_shape.dim[t_b_index].size

             bias_add_content = [0.] * bias_add_length

             bias_const_node = helper.create_constant_node(
                 bias_const_node_name, bias_add_content, dtypes.float32, shape=[bias_add_length])
             bias_node = helper.create_node('BiasAdd', bias_node_name, [target_node_name, bias_const_node_name])
             helper.set_attr_dtype(bias_node, "T", dtypes.float32)

             g.add_node(bias_node, target_node_name, next_node_names)
             g.add_node(bias_const_node, None, [bias_node_name])
             self.input_graph = g.dump_graph()
             self._parse_graph(self.input_graph)
             new_match_node_name=match_node_name[:1]+[bias_node_name]+match_node_name[1:]
             new_match_node_name=match_node_name[:1]+[bias_node_name]+match_node_name[1:]

             return new_match_node_name


    def _is_match(self, patterns):
        """Detect the rule matched nodes collections.

        Returns:
            [List] -- [the matched rule]
            [String] -- [the list contains the matched node name]
        """
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[0] for fusion in patterns):
                cur_node = self.node_name_mapping[list(
                    self.node_name_mapping.keys())[k]].node
                if cur_node.name != self.start_node_name:
                    continue

                if ((v in ("Conv2D", "DepthwiseConv2dNative")
                     and not self.enable_s8)
                    ) and not self._find_relu_node(cur_node):
                    continue
               
                for sub_rule in patterns:
                    if v != sub_rule[0]:
                        continue

                    sub_rule_len = len(sub_rule)
                    check_hardswish = True if sub_rule_len > 4  else False
                    self.logger.debug("Try to apply rule: {}".format(sub_rule))

                    cur_node_name = list(self.node_name_mapping.keys())[k]

                    matched_node_name.clear()

                    matched_node_name.append(cur_node_name)
                    count = 0
                    while sub_rule_len > 1:
                        if not self.node_name_mapping[cur_node_name].output:
                            self.logger.debug("Fail to match {}".format(sub_rule))
                            break

                        next_node_name = self.node_name_mapping[
                            cur_node_name].output[0]

                        is_shared_output = True if len(
                            self.node_name_mapping[cur_node_name].output
                        ) > 1 else False

                        add_op_quantizable = True 
                        is_hardswish = False
                        if is_shared_output:  # pragma: no cover
                            if next_node_name.find('hard_swish') != -1:
                                self.logger.debug("Find Hard Swish pattern ......")
                                is_hardswish = True
                                count = count + 1
                                if next_node_name.find('add') == -1:
                                    next_node_name = self.node_name_mapping[
                                    cur_node_name].output[1]
                            else:
                                add_op_quantizable = False
                        next_node_op = self.node_name_mapping[
                            next_node_name].node.op

                        if next_node_op in ("Add", "AddV2", "AddN"):
                            next_node = self.node_name_mapping[
                                next_node_name].node
                            next_node_inputs = list(next_node.input)
                            cur_node_index = next_node_inputs.index(
                                cur_node_name)

                            for index, input_name in enumerate(
                                    next_node_inputs):
                                node_type = self.node_name_mapping[helper.node_name_from_input(
                                    input_name)].node.op
                                if input_name != cur_node_name and index < cur_node_index and \
                                        node_type != 'Dequantize':
                                    add_op_quantizable = False
                                    break

                        if add_op_quantizable and next_node_op == sub_rule[1 - sub_rule_len]:
                            if not is_shared_output: 
                                matched_node_name.append(next_node_name)
                                sub_rule_len -= 1
                                cur_node_name = next_node_name
                            elif is_hardswish and self.new_api: # pragma: no cover
                                matched_node_name.append(next_node_name)
                                sub_rule_len -= 1
                                cur_node_name = next_node_name
                            else:
                                matched_node_name.clear()
                                self.logger.debug("Fail to match {}.".format(sub_rule))
                                break
                        else:
                            matched_node_name.clear()
                            self.logger.debug("Fail to match {}.".format(sub_rule))
                            break

                    if sub_rule_len == 1:
                         if check_hardswish and sub_rule[-1] == 'Mul' and \
                            sub_rule[-2] == 'Mul' and sub_rule[-3] == 'Relu6' and \
                            sub_rule[-4] == 'Add' and count != 1:
                              matched_node_name.clear()
                              self.logger.debug("Fail to match {}.".format(sub_rule))
                              break
                         self.logger.debug("Match {} on nodes {}.".
                                          format(sub_rule, matched_node_name))
                         return sub_rule, matched_node_name

        return None, None

    def _need_to_check(self, node_type):
        op_list = ("ConcatV2", "Conv2D", "Conv3D", "DepthwiseConv2D", "QuantizeV2", "DepthwiseConv2dNative",
                   "MaxPool", "MaxPool3D", "FusedBatchNormV3", "Requantize", "RequantizePerChannel", "AvgPool", "Pad",
                   "CropAndResize", "Dequantize", "Mean", "MatMul", "FakeQuantWithMinMaxVars")
        return any([node_type.find(i) != -1 for i in op_list])

    def _find_relu_node(self, node):
        #if node.op.find("HardSwish") != -1:
        #    return False
        if node.op in ("Relu", "Relu6") or \
            (node.op.find("AndRelu") != -1 and \
            ('alpha' not in node.attr or ('alpha' in node.attr and node.attr['alpha'].f == 0))):
            return True
        elif 'T' in node.attr and node.attr['T'].type in (dtypes.quint8, dtypes.uint8):
            return True
        elif (node.op.find("QuantizedConv") != -1
              or node.op.find("QuantizedDepthwiseConv") != -1 or
              node.op.find("QuantizedMatMul") != -1
              ) and (node.op.find("Relu") == -1 or \
              ('alpha' in node.attr and node.attr['alpha'].f > 0)):
            return False
        elif self._need_to_check(node.op):
            input_node = self.node_name_mapping[helper.node_name_from_input(
                node.input[0])]
            return self._find_relu_node(input_node.node)
        else:
            return False

    def _reset_output_node_maps(self):
        self.output_node_maps = {}

    def _get_op_list(self):
        self.op_list = []
        for _, v in enumerate(self.node_name_mapping):
            self.op_list.append(self.node_name_mapping[v].node.op)

    def _get_node_input(self, node_name):
        """
        Return control_input name, non-control_input node name
        """

        return [
            i for i in self.node_name_mapping[node_name].node.input
            if i[0] == '^'
        ], [
            i for i in self.node_name_mapping[node_name].node.input
            if i[0] != '^'
        ]

    def _intel_cpu_add_dequantize_result_node(self,
                                              quantized_output_name,
                                              original_node_name,
                                              dtype=dtypes.quint8,
                                              min_tensor_index=1):
        min_max_inputs = [
            "%s:%s" % (quantized_output_name, min_tensor_index),
            "%s:%s" % (quantized_output_name, min_tensor_index + 1)
        ]
        dequantize_name = original_node_name

        dequantize_node = helper.create_node(
            "Dequantize", dequantize_name,
            [quantized_output_name, min_max_inputs[0], min_max_inputs[1]])
        helper.set_attr_dtype(dequantize_node, "T", dtype)
        helper.set_attr_string(dequantize_node, "mode",
                               b"MIN_FIRST" if self.is_asymmetric else b"SCALED")
        self.add_output_graph_node(dequantize_node)

    def eightbitize_single_input_tensor_node(self, original_node,
                                             add_op_function):
        quantized_op_name = original_node.name + "_eightbit_quantized"
        quantized_op_type = "Quantized" + original_node.op
        if version1_gt_version2(tf.version.VERSION, '2.7.0') and original_node.op == "MaxPool3D":
            quantized_op_type = "_Quantized" + original_node.op
        all_input_names = self._add_eightbit_prologue_nodes(original_node.name)
        quantized_op_node = helper.create_node(quantized_op_type,
                                               quantized_op_name,
                                               all_input_names)
        add_op_function(original_node, quantized_op_node)
        self.add_output_graph_node(quantized_op_node)
        deq_type = dtypes.quint8 if self._find_relu_node(original_node) else dtypes.qint8
        self._intel_cpu_add_dequantize_result_node(quantized_op_name,
                                                   original_node.name,
                                                   dtype=deq_type)

    def _add_eightbit_prologue_nodes(self, original_node):
        namespace_prefix = original_node + "_eightbit"
        reshape_dims_name, reduction_dims_name = self._add_common_quantization_nodes(
            namespace_prefix, helper.node_name_from_input(
                self.node_name_mapping[original_node].node.input[0]))
        input_names = []
        min_max_names = []
        for each_input_name in self.node_name_mapping[original_node].node.input[:1]:
            if each_input_name[0] == '^':
                continue

            if self.node_name_mapping[original_node].node.op == "MatMul":
                # mkl ops _MklQuantizedMatMulWithBiasAndRelu|AndRequantize
                # requires the T1 data type as quint8
                dtype = dtypes.quint8
            else:
                input_node_name = helper.node_name_from_input(each_input_name)
                if input_node_name in self.output_node_maps:
                    if self.output_node_maps[input_node_name].op == "Dequantize":
                        dtype = dtypes.DType(
                            self.output_node_maps[input_node_name].attr["T"].type)
                    elif self._find_relu_node(self.node_name_mapping[original_node].node):
                        dtype = dtypes.quint8
                    else:
                        dtype = dtypes.qint8
                else:
                    dtype = dtypes.quint8 if self._find_relu_node(
                        self.node_name_mapping[original_node].node
                    ) else dtypes.qint8
                    if 'FusedBatchNorm' in self.node_name_mapping[original_node].node.op:
                        dtype = dtypes.qint8
            quantize_input_name, min_input_name, max_input_name = (
                self._eightbitize_input_to_node(namespace_prefix,
                                                each_input_name,
                                                reshape_dims_name,
                                                reduction_dims_name,
                                                dtype=dtype))
            input_names.append(quantize_input_name)
            min_max_names.append(min_input_name)
            min_max_names.append(max_input_name)
        all_input_names = []
        all_input_names.extend(input_names)
        if min_max_names:
            all_input_names.extend(min_max_names)

        for original_input_name in self.node_name_mapping[
                original_node].node.input:
            if original_input_name[0] == '^':
                all_input_names.append(original_input_name)
        return all_input_names

    def _add_common_quantization_nodes(self,
                                       namespace_prefix,
                                       control_input_names=None):
        """Builds constant nodes needed for quantization of inputs."""
        reshape_dims_name = namespace_prefix + "_reshape_dims"
        reduction_dims_name = namespace_prefix + "_reduction_dims"

        reshape_dims_node = helper.create_constant_node(
            reshape_dims_name, -1, dtypes.int32, [1])

        self.add_output_graph_node(reshape_dims_node)
        reduction_dims_node = helper.create_constant_node(
            reduction_dims_name, 0, dtypes.int32, [1])

        self.add_output_graph_node(reduction_dims_node)
        return reshape_dims_name, reduction_dims_name

    def add_output_graph_node(self, output_node):
        """Inserts one node into the new graph."""
        assert output_node.name not in self.output_node_maps
        self.output_node_maps[output_node.name] = output_node
        self.output_graph.node.extend([output_node])

    def _parse_graph(self, input_graph=None):
        """
        Parse the graph and get the input node and output node name details.
        """
        self.logger.debug("Start to parse graph.")

        graph = self.input_graph if input_graph is None else input_graph
        self.node_name_mapping = {}

        for node in graph.node:
            # each_node = self.node_details(node=node,  output=[])
            if node.name in self.node_name_mapping:
                raise ValueError(
                    "Duplicate Node Found when _parse_graph, the node name is {}" .format(
                        node.name))
            self.node_name_mapping[node.name] = self.node_details(node=node, output=[])
        for node_name in self.node_name_mapping:
            for each_input in self.node_name_mapping[node_name].node.input:
                self.node_name_mapping[helper.node_name_from_input(each_input)].output.\
                    append(node_name)

    def remove_redundant_quantization(self, old_graph):
        old_nodes_map = self.create_nodes_map(old_graph)
        self.output_graph = graph_pb2.GraphDef()
        inputs_to_rename = {}
        # We go through all the nodes, looking for any that match the patterns we
        # know how to optimize away.
        for node in old_graph.node:
            # We always start with a Quantize node, and examine its inputs to see if
            # they are in a form that can be removed.
            if node.op not in ["Quantize", "QuantizeV2"]:
                continue

            dequantize_node_name = helper.node_name_from_input(node.input[0])

            assert dequantize_node_name in old_nodes_map, "Input node name '" + \
                dequantize_node_name + "' not found in node '" + node.name + "'"

            dequantize_node = old_nodes_map[dequantize_node_name]
            # Do we have a Dequantize feeding in, with the same type as the
            # Quantize?
            if dequantize_node.op != "Dequantize":
                continue

            if node.attr["T"] != dequantize_node.attr["T"]:
                continue

            # Now look at the other inputs, and ensure they're Min/Max nodes.
            min_node_name = helper.node_name_from_input(node.input[1])
            max_node_name = helper.node_name_from_input(node.input[2])
            min_node = old_nodes_map[min_node_name]
            max_node = old_nodes_map[max_node_name]
            is_min_right_type = (min_node.op in ["Min", "Dequantize"])
            is_max_right_type = (max_node.op in ["Max", "Dequantize"])
            if not is_min_right_type or not is_max_right_type:
                self.logger.info("Not find expected types on inputs {}, {}.".
                                 format(min_node.op, max_node.op))
                continue
            min_node_input_name = helper.node_name_from_input(
                min_node.input[0])
            max_node_input_name = helper.node_name_from_input(
                max_node.input[0])
            # There are two different patterns for Min nodes we can recognize, one
            # where the input comes directly from the same one as the Max, and
            # another where we run it through another Min first, so check for
            # both.
            is_same_input = False
            if min_node_input_name == max_node_input_name:
                is_same_input = True
            else:
                first_min_node_input = old_nodes_map[min_node_input_name]
                if first_min_node_input.op == "Concat":
                    second_min_node_name = helper.node_name_from_input(
                        first_min_node_input.input[1])
                    second_min_node = old_nodes_map[second_min_node_name]
                    if second_min_node.op == "Min":
                        second_min_node_input_name = helper.node_name_from_input(
                            second_min_node.input[0])
                        is_same_input = (
                            second_min_node_input_name == max_node_input_name)
            if not is_same_input:
                self.logger.info("Different min/max inputs {}.".format(min_node_input_name))
                continue
            # We recognize this pattern, so mark the graph edges to be rewired to
            # route around it entirely, since we know it's a no-op.
            dequantize_source_name = helper.node_name_from_input(
                dequantize_node.input[0])
            node_tensor_name = helper.ensure_tensor_name_has_port(node.name)
            min_tensor_name = node.name + ":1"
            max_tensor_name = node.name + ":2"

            inputs_to_rename[node_tensor_name] = dequantize_source_name
            inputs_to_rename[min_tensor_name] = dequantize_node.input[1]
            inputs_to_rename[max_tensor_name] = dequantize_node.input[2]
        # Finally we apply all the rewiring we've marked to the graph.
        for node in old_graph.node:
            for index, input_full_name in enumerate(node.input):
                input_name = helper.ensure_tensor_name_has_port(
                    input_full_name)
                if input_name in inputs_to_rename:
                    node.input[index] = inputs_to_rename[input_name]
            self.add_output_graph_node(node)
        return self.output_graph

    def create_nodes_map(self, graph):
        """Builds a mapping of node names to their defs from the graph."""
        nodes_map = {}
        for node in graph.node:
            assert node.name not in nodes_map, "Duplicate node names detected."
            nodes_map[node.name] = node

        return nodes_map

    def _add_quantize_down_nodes(self,
                                 original_node,
                                 quantized_output_name,
                                 requantize_type=dtypes.quint8,
                                 is_relu6=False):
        quantized_outputs = [
            quantized_output_name, quantized_output_name + ":1",
            quantized_output_name + ":2"
        ]
        if not self.fake_quant:
            # Add a RequantizationRange node for finding the min and max values.
            requant_range_node = helper.create_node(
                "RequantizationRangePerChannel"
                if self.per_channel else "RequantizationRange",
                original_node.name + "_eightbit_requant_range", quantized_outputs)

            if self.per_channel:
                helper.set_attr_dtype(requant_range_node, "T", dtypes.qint32)
                if is_relu6:
                    helper.set_attr_float(requant_range_node, "clip_value_max",
                                          6.0)
                else:
                    helper.set_attr_float(requant_range_node, "clip_value_max",
                                          1e30)
            else:
                helper.set_attr_dtype(requant_range_node, "Tinput", dtypes.qint32)

            self.add_output_graph_node(requant_range_node)
            min_max_inputs = [
                requant_range_node.name + ":0", requant_range_node.name + ":1"
            ]
        else:
            max_input_name = original_node.name + "_max"
            max_node = helper.create_constant_node(
                max_input_name, 1., dtypes.float32)
            self.add_output_graph_node(max_node)

            min_input_name = original_node.name + "_min"
            min_node = helper.create_constant_node(
                min_input_name, -1., dtypes.float32)
            self.add_output_graph_node(min_node)

            min_max_inputs = [
                min_input_name, max_input_name
            ]

        requantize_node = helper.create_node(
            "RequantizePerChannel" if self.per_channel else "Requantize",
            original_node.name + "_eightbit_requantize",
            quantized_outputs + min_max_inputs)
        if self.per_channel:
            helper.set_attr_dtype(requantize_node, "T", dtypes.qint32)
        else:
            helper.set_attr_dtype(requantize_node, "Tinput", dtypes.qint32)

        helper.set_attr_dtype(requantize_node, "out_type", requantize_type)
        self.add_output_graph_node(requantize_node)
        return requantize_node.name

    def _eightbitize_input_to_node(self,
                                   namespace_prefix,
                                   original_input_name,
                                   reshape_dims_name,
                                   reduction_dims_name,
                                   dtype=dtypes.quint8):
        """Takes one float input to an op, and converts it to quantized form."""
        unique_input_name = helper.unique_node_name_from_input(
            original_input_name)
        if unique_input_name in self.quantized_node_dict:
            quantized_tuple = self.quantized_node_dict[unique_input_name]
            return quantized_tuple[0], quantized_tuple[1], quantized_tuple[2]

        if self.fake_quant:
            min_input_name = namespace_prefix + "_min_" + unique_input_name
            min_node = helper.create_constant_node(
                min_input_name, -1., dtypes.float32)
            self.add_output_graph_node(min_node)
            max_input_name = namespace_prefix + "_max_" + unique_input_name
            max_node = helper.create_constant_node(
                max_input_name, 1., dtypes.float32)
            self.add_output_graph_node(max_node)
            quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name
            quantize_input_node = helper.create_node(
                "QuantizeV2", quantize_input_name,
                [original_input_name, min_input_name, max_input_name])
        else:
            reshape_input_name = namespace_prefix + "_reshape_" + unique_input_name
            min_input_name = namespace_prefix + "_min_" + unique_input_name
            max_input_name = namespace_prefix + "_max_" + unique_input_name
            quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name
            reshape_input_node = helper.create_node(
                "Reshape", reshape_input_name,
                [original_input_name, reshape_dims_name])
            helper.set_attr_dtype(reshape_input_node, "T", dtypes.float32)
            self.add_output_graph_node(reshape_input_node)
            min_input_node = helper.create_node(
                "Min", min_input_name, [reshape_input_name, reduction_dims_name])
            helper.set_attr_dtype(min_input_node, "T", dtypes.float32)
            helper.set_attr_dtype(min_input_node, "Tidx", dtypes.int32)
            helper.set_attr_bool(min_input_node, "keep_dims", False)
            self.add_output_graph_node(min_input_node)
            max_input_node = helper.create_node(
                "Max", max_input_name, [reshape_input_name, reduction_dims_name])
            helper.set_attr_dtype(max_input_node, "T", dtypes.float32)
            helper.set_attr_dtype(max_input_node, "Tidx", dtypes.int32)
            helper.set_attr_bool(max_input_node, "keep_dims", False)
            self.add_output_graph_node(max_input_node)
            quantize_input_node = helper.create_node(
                "QuantizeV2", quantize_input_name,
                [original_input_name, min_input_name, max_input_name])

        helper.set_attr_dtype(quantize_input_node, "T", dtype)
        helper.set_attr_string(quantize_input_node, "mode",
                               b"MIN_FIRST" if self.is_asymmetric else b"SCALED")
        if not self.is_asymmetric:
            helper.set_attr_string(quantize_input_node, "round_mode", b"HALF_TO_EVEN")
        # if FLAGS.model_name in ["wide_deep_large_ds"]:
        #    set_attr_string(quantize_input_node, "mode", b"MIN_FIRST")
        # else:
        #    set_attr_string(quantize_input_node, "mode",
        #                    b"SCALED" if self.intel_cpu_eightbitize else b"MIN_FIRST")
        #    set_attr_string(quantize_input_node, "round_mode",
        #                    b"HALF_TO_EVEN" if self.intel_cpu_eightbitize
        #                    else b"HALF_AWAY_FROM_ZERO")
        self.add_output_graph_node(quantize_input_node)
        min_output_name = quantize_input_name + ":1"
        max_output_name = quantize_input_name + ":2"
        self.quantized_node_dict[unique_input_name] = (quantize_input_name,
                                                       min_output_name,
                                                       max_output_name)
        return quantize_input_name, min_output_name, max_output_name


    def _intel_cpu_quantize_weight_eightbit(self,
                                            parent,
                                            input_node,
                                            per_channel):
        qint8_const_node, min_node, max_node = helper.generate_quantized_weight_node(
            parent, input_node, per_channel, self.weight_bit, self.device)
        self.add_output_graph_node(qint8_const_node)
        self.add_output_graph_node(min_node)
        self.add_output_graph_node(max_node)

        return qint8_const_node.name, min_node.name, max_node.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce

def deep_get(dictionary, keys, default=None):

    """get the dot key's item in nested dict
       eg person = {'person':{'name':{'first':'John'}}}
       deep_get(person, "person.name.first") will output 'John'

       Args:
           dictionary (dict): The dict object to get keys
           keys (dict): The deep keys
           default (object): The return item if key not exists
       Returns:
           item: the item of the deep dot keys
    """
    return reduce(lambda d, key: d.get(key, default) \
        if isinstance(d, dict) else default, keys.split("."), dictionary)

def deep_set(dictionary, keys, value):

    """set the dot key's item in nested dict
       eg person = {'person':{'name':{'first':'John'}}}
       deep_set(person, "person.sex", 'male') will output
       {'person': {'name': {'first': 'John'}, 'sex': 'male'}}

       Args:
           dictionary (dict): The dict object to get keys
           keys (dict): The deep keys
           value (object): The value of the setting key
    """
    keys = keys.split('.')
    for key in keys[:-1]:
        dictionary = dictionary.setdefault(key, {})
    dictionary[keys[-1]] = value

class DotDict(dict):
    """access yaml using attributes instead of using the dictionary notation.

    Args:
        value (dict): The dict object to access.

    """

    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError('expected dict')

    def __getitem__(self, key):
        value = self.get(key, None)
        return value

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, DotDict):
            value = DotDict(value)
        if isinstance(value, list) and len(value) == 1 and isinstance(
                value[0], dict):
            value = DotDict(value[0])
        if isinstance(value, list) and len(value) > 1 and all(isinstance(
                v, dict) for v in value):
            value = DotDict({k: v for d in value for k, v in d.items()})
        super(DotDict, self).__setitem__(key, value)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    __setattr__, __getattr__ = __setitem__, __getitem__




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
from neural_compressor.utils.utility import dump_elapsed_time
from .fuse_column_wise_mul import FuseColumnWiseMulOptimizer
from .remove_training_nodes import RemoveTrainingNodesOptimizer
from .split_shared_input import SplitSharedInputOptimizer
from .strip_unused_nodes import StripUnusedNodesOptimizer
from .graph_cse_optimizer import GraphCseOptimizer
from .fold_constant import GraphFoldConstantOptimizer
from .fold_batch_norm import FoldBatchNormNodesOptimizer
from .rename_batch_norm import RenameBatchNormOptimizer
from .update_enter import UpdateEnterOptimizer
from .convert_layout import ConvertLayoutOptimizer
from .fuse_gelu import FuseGeluOptimizer
from .fuse_reshape_transpose import FuseTransposeReshapeOptimizer
from .convert_leakyrelu import ConvertLeakyReluOptimizer
from .dummy_biasadd import InjectDummyBiasAddOptimizer
from .convert_add_to_biasadd import ConvertAddToBiasAddOptimizer
from .grappler_pass import GrapplerOptimizer
from .fuse_conv_with_math import FuseConvWithMathOptimizer
from .fuse_biasadd_add import FuseBiasAddAndAddOptimizer
from .switch_optimizer import SwitchOptimizer
from .move_squeeze_after_relu import MoveSqueezeAfterReluOptimizer
from .convert_nan_to_random import ConvertNanToRandom
from .expanddims_optimizer import ExpandDimsOptimizer
from .fetch_weight_from_reshape import FetchWeightFromReshapeOptimizer
from .fuse_decomposed_bn import FuseDecomposedBNOptimizer


class PreOptimization():
    def __init__(self, model, optimization, new_api):
        self.model = model
        self.optimization = optimization
        self.new_api = new_api

        self.analyzer = GraphAnalyzer()
        self.analyzer.graph = model.graph_def
        self.analyzer.parse_graph()
        self._tmp_graph_def = None
        self._excluded_node_names = []


    def get_excluded_node_names(self):
        """Get the excluded node name

        Returns:
            string list: the excluded ops' name
        """
        return self._excluded_node_names

    @dump_elapsed_time("Pass Pre Optimization")
    def get_optimized_model(self, itex_mode=False):
        """Executed the non-precision dependant graph optimization.
        The input graph will be optimized with following passes:
        1. Remove the training nodes like Identity Op.
        2. Split the shared nodes like weights node for multi-Conv2d.
        3. Fold Constant Nodes as less as possible.
        4. Fuse the Mul node into the previous Conv2D/MatMul if possible.
        5. Strip the useless nodes.
        6. Do the Common sequence elimation optimization on the graph.
        7. Fold the BN node into the previous Conv2D if possible.

        Returns:
            [graphdef]: the optimized graphdef object.
        """

        from neural_compressor.experimental.common import Model

        origin_model = Model(self.model._model, **self.model.kwargs)
        origin_model.name = self.model.name
        origin_model.model_type = self.model.model_type
        origin_model.output_tensor_names = self.model.output_tensor_names
        origin_model.input_tensor_names = self.model.input_tensor_names
        origin_model.workspace_path = self.model.workspace_path

        output_node_names = self.model.output_node_names
        input_node_names = self.model.input_node_names
        input_output_names = output_node_names + input_node_names

        self._tmp_graph_def = ConvertLayoutOptimizer(
            self.model.graph_def, output_node_names).do_transformation()

        self._tmp_graph_def = GrapplerOptimizer(
            self._tmp_graph_def, input_output_names, self.optimization).do_transformation()

        self._tmp_graph_def = SwitchOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = RemoveTrainingNodesOptimizer(
            self._tmp_graph_def, protected_nodes=input_output_names).do_transformation()

        self._tmp_graph_def = SplitSharedInputOptimizer(self._tmp_graph_def).do_transformation()

        # disable fold constant for itex qdq mode
        if not itex_mode:
            self._tmp_graph_def = GraphFoldConstantOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseColumnWiseMulOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripUnusedNodesOptimizer(self._tmp_graph_def,
            input_node_names, output_node_names).do_transformation()

        self._tmp_graph_def = FuseGeluOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = GraphCseOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseDecomposedBNOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FoldBatchNormNodesOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = RenameBatchNormOptimizer(
            self._tmp_graph_def).do_transformation()

        #TODO we should handle all control ops elegantly not bypass it.
        self._tmp_graph_def, excluded_node_names = UpdateEnterOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertLeakyReluOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertAddToBiasAddOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseTransposeReshapeOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseConvWithMathOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ExpandDimsOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FetchWeightFromReshapeOptimizer(
            self._tmp_graph_def).do_transformation()
        if not self.new_api:
            #TODO we need to remove below optimizer once the TF enabled the single
            # matmul op quantization
            self._tmp_graph_def = InjectDummyBiasAddOptimizer(
                self._tmp_graph_def, output_node_names).do_transformation()
        self._tmp_graph_def = FuseBiasAddAndAddOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = MoveSqueezeAfterReluOptimizer(
            self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertNanToRandom(
            self._tmp_graph_def).do_transformation()

        self._excluded_node_names.extend(excluded_node_names)
        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)

        origin_model.graph_def = self._tmp_graph_def

        return origin_model

    def get_matched_nodes(self, patterns):
        """Searche the matched nodes with the specified patterns

        Args:
            patterns ([string list]): The patterns should be illustrated as below.
                [['MatMul'], ("BiasAdd"), ("Relu",)]
        Returns:
            [string list]: It will return the list that contains the matched nodes name
                and pattern. ['matched_node_a_name', 'matched_node_a_name',['MatMul','BiasAdd']]
        """
        self.analyzer.graph = self._tmp_graph_def
        self.analyzer.parse_graph()
        res = []

        for sub_pattern in patterns:
            res.extend([i for i in self.analyzer.query_fusion_pattern_nodes(
                sub_pattern) if i not in res])
        return res

    def has_positive_input(self, node_name):
        """Check the specified node has the positive input or not.

        Args:
            node_name ([string]): node's name

        Returns:
            [bool]: True if the node has the positive input data,
                    False if the node has the negative input data.
        """
        return self.analyzer.has_positive_input(node_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np
import glob
from .dataset import dataset_registry, Dataset


@dataset_registry(dataset_type="style_transfer", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class StyleTransferDataset(Dataset):
    """Dataset used for style transfer task.
       This Dataset is to construct a dataset from two specific image holders representing
       content image folder and style image folder.

    Args: content_folder (str):Root directory of content images.
          style_folder (str):Root directory of style images.
          crop_ratio (float, default=0.1):cropped ratio to each side.
          resize_shape (tuple, default=(256, 256)):target size of image.
          image_format (str, default='jpg'): target image format.
          transform (transform object, default=None):  transform to process input data.
          filter (Filter objects, default=None): filter out examples according 
                                                 to specific conditions
    """

    def __init__(self, content_folder, style_folder, crop_ratio=0.1,
                 resize_shape=(256, 256), image_format='jpg', transform=None, filter=None):

        self.transform = transform
        self.content_folder = content_folder
        self.style_folder = style_folder
        self.resize_shape = resize_shape
        self.crop_ratio = crop_ratio
        self.content_images = glob.glob(os.path.join(content_folder, '*' + image_format))
        self.style_images = glob.glob(os.path.join(style_folder, '*' + image_format))
        self.image_list = []
        for content in self.content_images:
            for style in self.style_images:
                self.image_list.append((content, style))

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, index):
        from PIL import Image
        content_image, style_image = self.image_list[index]
        content_image = Image.open(content_image)
        style_image = Image.open(style_image)
        width, height = style_image.size
        crop_ratio = self.crop_ratio
        crop_box = (
            crop_ratio * height,
            crop_ratio * width,
            (1 - crop_ratio) * height,
            (1 - crop_ratio) * width)
        content_image = np.asarray(content_image.resize(self.resize_shape))
        style_image = np.asarray(style_image.resize(self.resize_shape))
        if content_image.max() > 1.0:
            content_image = content_image / 255.
        if style_image.max() > 1.0:
            style_image = style_image / 255.

        return (content_image, style_image), 0



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import subprocess
import logging
import time

from . import globals

if not os.path.exists("neural_coder_workspace"):
    os.makedirs("neural_coder_workspace")


def enable(
    code,
    features,
    target_batch_size=1,  # effective for feature "pytorch_change_batch_size"
    num_benchmark_iteration=30,  # effective for feature "pytorch_benchmark"
    generate_patch=True,
    overwrite=False,
    # TO-ADD: return a folder with user-input code and artificial pip folders
    brutal_mode=False,
    save_patch_path="",
    patch_suffix=".diff",
    remove_copy=True,
    consider_imports=True,
    patch_imports=False,
    logging_level="info",
    run_bench=False,
    entry_code="",
    entry_code_args="",
    mode="throughput",
    cpu_set_env=True,
    ncore_per_instance=-1,  # only for "self_defined" mode
    ninstances=-1,  # only for "self_defined" mode
    bench_batch_size=-1,  # only for "self_defined" mode
):

    # Preparation

    # set up workspace
    ws_path = "neural_coder_workspace/" + \
        "enable" + str(int(time.time())) + "/"
    os.makedirs(ws_path)

    # user parameters
    globals.consider_imports = consider_imports
    logging_var = "logging." + logging_level.upper()
    globals.logging_level = eval(logging_var)

    # set up logging
    logger = logging.getLogger(ws_path)
    logger.setLevel(globals.logging_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s: - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

    fh = logging.FileHandler(ws_path+'enable.log')
    fh.setLevel(globals.logging_level)
    fh.setFormatter(formatter)
    ch = logging.StreamHandler()
    ch.setLevel(globals.logging_level)
    ch.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)

    # print key inputs
    logger.info(f"Enabling started ...")
    logger.info(f"code: {code}")
    logger.info(f"features: {features}")

    # feature list for reference
    '''
    feature_list = ["pytorch_jit_script",
                    "pytorch_jit_script_ofi",
                    "pytorch_inc_dynamic_quant",
                    "pytorch_inc_static_quant",
                    "pytorch_ipex_fp32",
                    "pytorch_ipex_bf16",
                    "pytorch_ipex_int8_static_quant",
                    "pytorch_ipex_int8_dynamic_quant",
                    "pytorch_channels_last",
                    "pytorch_mixed_precision_cpu",
                    "pytorch_mixed_precision_cuda",
                    "pytorch_torchdynamo_jit_script",
                    "pytorch_torchdynamo_jit_script_ofi",
                    "pytorch_torchdynamo_jit_trace",
                    "pytorch_torchdynamo_jit_trace_ofi",
                    "pytorch_benchmark",
                    "pytorch_change_batch_size",
                    "pytorch_cuda_to_cpu",
                    "pytorch_lightning_bf16_cpu",
                    "tensorflow_amp",
                    "keras_amp",]
    '''

    # Benchmark
    if run_bench:
        # add "pytorch_change_batch_size" to features
        from .utils.cpu_info import get_num_cpu_cores
        ncores = get_num_cpu_cores()
        if mode == "throughput":
            target_batch_size = 2 * ncores
        elif mode == "multi_instance":
            target_batch_size = 1
        elif mode == "latency":
            target_batch_size = 1
        elif mode == "self_defined":
            target_batch_size = bench_batch_size

        if "pytorch_change_batch_size" not in features:
            features.append("pytorch_change_batch_size")

        # add "pytorch_benchmark" to features
        if "pytorch_benchmark" not in features:
            features.append("pytorch_benchmark")

        logger.info(
            f"Will perform benchmark on [{mode}] mode with batch size [{target_batch_size}] ...")

    # Rearrange Feature Order (due to certain regulations: e.g. channels_last is before IPEX, and IPEX is before JIT)
    from .utils.common import move_element_to_front
    for feature in ["pytorch_benchmark",
                    "pytorch_channels_last",
                    "pytorch_ipex_fp32",
                    "pytorch_ipex_bf16",
                    "pytorch_ipex_int8_static_quant",
                    "pytorch_ipex_int8_dynamic_quant",
                    "pytorch_jit_script",
                    "pytorch_jit_script_ofi",
                    "pytorch_torchdynamo_jit_script",
                    "pytorch_torchdynamo_jit_script_ofi",
                    "pytorch_torchdynamo_jit_trace",
                    "pytorch_torchdynamo_jit_trace_ofi",
                    "pytorch_inc_static_quant",
                    "pytorch_inc_dynamic_quant",
                    "pytorch_mixed_precision_cpu",
                    "pytorch_mixed_precision_cuda", ]:
        features = move_element_to_front(features, feature)

    # Enabling

    transformed_list_code_path = []

    for feature in features:

        # reset globals
        globals.reset_globals()

        from .utils import handle_user_input
        globals.list_code_path, num_user_code_path = handle_user_input.get_all_code_path(
            code)
        if len(transformed_list_code_path) > 0:
            globals.list_code_path = transformed_list_code_path

        # common for all features (transformations),
        list_transformed_code = []
        # in this list, each item stores the transformed code of the corresponding original code
        # by the order in code_path

        # global behaviors
        logger.info(
            f"Performing code transformation for feature: [{feature}] ...")

        from .graphers.code_line import register_code_line
        from .graphers.model import register_nnModule_class, register_nnModule_instance_definition
        from .graphers.function import register_func_wrap_pair
        from .coders.transform import execute_insert_transformation, execute_indenting_transformation

        register_code_line()
        register_func_wrap_pair()
        register_nnModule_class()
        register_nnModule_instance_definition()

        # AMP
        if "pytorch_mixed_precision_cpu" == feature:
            from .coders.pytorch.amp import PTAMP
            opt = PTAMP("cpu")
            opt.register_transformation()
        elif "pytorch_mixed_precision_cuda" == feature:
            from .coders.pytorch.amp import PTAMP
            opt = PTAMP("cuda")
            opt.register_transformation()

        # TorchDynamo
        if "pytorch_torchdynamo_jit_script" == feature:
            from .coders.pytorch.torchdynamo import TorchDynamo
            opt = TorchDynamo()
            opt.register_transformation()
            from .coders.pytorch.torchdynamo import TorchDynamoJITScript
            opt = TorchDynamoJITScript(
                globals.list_model_def_instance, "plain")
            opt.register_transformation()
        elif "pytorch_torchdynamo_jit_script_ofi" == feature:
            from .coders.pytorch.torchdynamo import TorchDynamo
            opt = TorchDynamo()
            opt.register_transformation()
            from .coders.pytorch.torchdynamo import TorchDynamoJITScript
            opt = TorchDynamoJITScript(globals.list_model_def_instance, "ofi")
            opt.register_transformation()
        elif "pytorch_torchdynamo_jit_trace" == feature:
            from .coders.pytorch.torchdynamo import TorchDynamo
            opt = TorchDynamo()
            opt.register_transformation()
            from .coders.pytorch.torchdynamo import TorchDynamoJITTrace
            opt = TorchDynamoJITTrace(globals.list_model_def_instance, "plain")
            opt.register_transformation()
        elif "pytorch_torchdynamo_jit_trace_ofi" == feature:
            from .coders.pytorch.torchdynamo import TorchDynamo
            opt = TorchDynamo()
            opt.register_transformation()
            from .coders.pytorch.torchdynamo import TorchDynamoJITTrace
            opt = TorchDynamoJITTrace(globals.list_model_def_instance, "ofi")
            opt.register_transformation()

        # Channels Last
        if "pytorch_channels_last" == feature:
            from .coders.pytorch.channels_last import ChannelsLast
            opt = ChannelsLast(globals.list_model_def_instance)
            opt.register_transformation()

        # JIT
        if "pytorch_jit_script" == feature:
            from .coders.pytorch.jit import JITScript
            opt = JITScript(globals.list_model_def_instance, "plain")
            opt.register_transformation()

        elif "pytorch_jit_script_ofi" == feature:
            from .coders.pytorch.jit import JITScript
            opt = JITScript(globals.list_model_def_instance, "ofi")
            opt.register_transformation()

        # IPEX
        if "pytorch_ipex_fp32" == feature:
            from .coders.pytorch.intel_extension_for_pytorch.ipex import IPEX
            opt = IPEX(globals.list_model_def_instance, "fp32")
            opt.register_transformation()

        elif "pytorch_ipex_bf16" == feature:
            from .coders.pytorch.intel_extension_for_pytorch.ipex import IPEX
            opt = IPEX(globals.list_model_def_instance, "bf16")
            opt.register_transformation()
            from .coders.pytorch.amp import PTAMP  # enable amp together
            opt = PTAMP("cpu")
            opt.register_transformation()

        elif "pytorch_ipex_int8_static_quant" == feature:
            from .coders.pytorch.intel_extension_for_pytorch.ipex import IPEX
            opt = IPEX(globals.list_model_def_instance, "int8_static_quant")
            opt.register_transformation()

        elif "pytorch_ipex_int8_dynamic_quant" == feature:
            from .coders.pytorch.intel_extension_for_pytorch.ipex import IPEX
            opt = IPEX(globals.list_model_def_instance, "int8_dynamic_quant")
            opt.register_transformation()

        # INC
        if "pytorch_inc_dynamic_quant" == feature:
            from .coders.pytorch.neural_compressor.dynamic_quant import DynamicQuant
            opt = DynamicQuant(globals.list_model_def_instance)
            opt.register_transformation()

        elif "pytorch_inc_static_quant" == feature:
            from .coders.pytorch.dummy_dataloader import DummyDataLoader  # detect dataloader first
            opt = DummyDataLoader(globals.list_model_def_instance)
            opt.register_transformation()
            from .coders.pytorch.neural_compressor.static_quant import StaticQuant
            opt = StaticQuant(globals.list_model_def_instance)
            opt.register_transformation()

        # Benchmark
        if "pytorch_benchmark" == feature:
            from .coders.pytorch.benchmark import Benchmark
            globals.num_benchmark_iteration = str(num_benchmark_iteration)
            opt = Benchmark()
            opt.register_transformation()

        # transformation execution
        for i in globals.list_code_path:
            list_transformed_code.append(open(i, 'r').read())
        list_transformed_code = execute_indenting_transformation(
            list_transformed_code)
        list_transformed_code = execute_insert_transformation(
            list_transformed_code)

        # other features (which use direct line transform instead of register-and-execute transform,
        # these features will be transformed here)
        for i in range(len(list_transformed_code)):
            # Batch Size
            if "pytorch_change_batch_size" == feature:
                from .coders.pytorch.batch_size import BatchSizeCoder
                globals.target_batch_size = str(target_batch_size)
                list_transformed_code[i] = BatchSizeCoder(
                    list_transformed_code[i]).transform()

            # CUDA to CPU
            if "pytorch_cuda_to_cpu" == feature:
                from .coders.pytorch.cuda_to_cpu import CudaToCpu
                list_transformed_code[i] = CudaToCpu(
                    list_transformed_code[i]).transform()

            # Lightning
            if "pytorch_lightning_bf16_cpu" == feature:
                from .coders.pytorch.lightning import Lightning
                list_transformed_code[i] = Lightning(
                    list_transformed_code[i]).transform()

            # TF & Keras AMP
            if "tensorflow_mixed_precision" == feature:
                from .coders.tensorflow.amp import TensorFlowKerasAMP
                list_transformed_code[i] = TensorFlowKerasAMP(
                    list_transformed_code[i]).transform()

        logger.info(f"Code transformation for feature: [{feature}] finished.")

        for path in globals.list_code_path:
            if path[-14:] == "_nc_enabled.py":
                path_transformed = path
            else:
                path_transformed = path[:-3] + "_nc_enabled.py"
            open(path_transformed, "w").write(
                list_transformed_code[globals.list_code_path.index(path)])
            globals.list_code_path[globals.list_code_path.index(
                path)] = path_transformed
        transformed_list_code_path = globals.list_code_path

    # Output of Enabling

    globals.list_code_path, num_user_code_path = handle_user_input.get_all_code_path(
        code)

    if save_patch_path == "":
        save_patch_path = ws_path

    if generate_patch:
        whole_patch_user_code = ""
        for path in globals.list_code_path[0:num_user_code_path]:
            path_transformed = path[:-3] + "_nc_enabled.py"
            cmd_gen_patch = "diff -up " + path + " " + path_transformed
            sp_gen_patch = subprocess.Popen(
                cmd_gen_patch, env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
            sp_gen_patch.wait()
            this_patch, _ = sp_gen_patch.communicate()
            this_patch = str(this_patch)[2:-1]
            whole_patch_user_code += this_patch
        open(save_patch_path + "neural_coder_patch" + patch_suffix, "w").write(
            whole_patch_user_code.replace(r'\n', '\n').replace(r'\t', '\t').replace(r"\'", "\'"))
        abs_patch_path = os.path.abspath(
            save_patch_path + "neural_coder_patch" + patch_suffix)
        logger.info(f"The patch is saved to: [{abs_patch_path}]")

        if overwrite:
            sp_overwrite = subprocess.Popen(
                "patch -d/ -p0 < " + abs_patch_path, env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
            sp_overwrite.wait()
            os.remove(abs_patch_path)  # remove patch after overwrite

        if patch_imports:
            whole_patch_import_modules = ""
            for path in globals.list_code_path[num_user_code_path:]:
                path_transformed = path[:-3] + "_nc_enabled.py"
                cmd_gen_patch = "diff -up " + path + " " + path_transformed
                sp_gen_patch = subprocess.Popen(
                    cmd_gen_patch, env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
                sp_gen_patch.wait()
                this_patch, _ = sp_gen_patch.communicate()
                this_patch = str(this_patch)[2:-1]
                whole_patch_import_modules += this_patch
            open(save_patch_path + "neural_coder_patch_import_modules" + patch_suffix, "w").write(
                whole_patch_import_modules.replace(r'\n', '\n').replace(r'\t', '\t').replace(r"\'", "\'"))
            abs_patch_path = os.path.abspath(
                save_patch_path + "neural_coder_patch_import_modules" + patch_suffix)
            logger.info(
                f"The patch for imported modules is saved to: [{abs_patch_path}]")

    # remove copy for imports
    if remove_copy:
        for path in globals.list_code_path:
            try:
                path_transformed = path[:-3] + "_nc_enabled.py"
                os.remove(path_transformed)
            except:
                pass

    # Benchmark
    if run_bench:
        bench_performance, bench_mode, bench_ws_path = bench(
            code=code,
            entry_code=entry_code,
            entry_code_args=entry_code_args,
            patch_path=abs_patch_path,
            mode=mode,
            cpu_set_env=cpu_set_env,
            ncore_per_instance=ncore_per_instance,  # only for "self_defined" mode
            ninstances=ninstances,  # only for "self_defined" mode
            bench_batch_size=bench_batch_size,  # only for "self_defined" mode
        )

        return bench_performance, bench_mode, bench_ws_path


'''
bench API works on either "optimized code", or "patch" + "original code"
it does not enable benchmark, or enable change of batch size, all the enabling is done in enable API
which means the "optimized code" should already have "pytorch_benchmark" and "pytorch_change_batch_size" enabled
or the "patch" should already have the code modification for "pytorch_benchmark" and 
"pytorch_change_batch_size" in it
'''


def bench(
    code,
    entry_code="",
    entry_code_args="",
    patch_path="",
    mode="throughput",  # throughput, latency, multi_instance or self_defined
    logging_level="info",
    cpu_set_env=True,
    ncore_per_instance=-1,  # only for "self_defined" mode
    ninstances=-1,  # only for "self_defined" mode
    bench_batch_size=-1,  # only for "self_defined" mode
):

    # set up workspace
    ws_path = "neural_coder_workspace/" + "bench" + str(int(time.time())) + "/"
    os.makedirs(ws_path)

    # set up logging
    logging_var = "logging." + logging_level.upper()
    globals.logging_level = eval(logging_var)

    logger = logging.getLogger(ws_path)
    logger.setLevel(globals.logging_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s: - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

    fh = logging.FileHandler(ws_path+'bench.log')
    fh.setLevel(globals.logging_level)
    fh.setFormatter(formatter)
    ch = logging.StreamHandler()
    ch.setLevel(globals.logging_level)
    ch.setFormatter(formatter)

    logger.addHandler(ch)
    logger.addHandler(fh)

    # print key inputs
    logger.info(f"Benchmarking started ...")
    logger.info(f"code: {code}")
    logger.info(f"mode: {mode}")

    # entry code
    if entry_code == "":
        # if not specify entry_code, then code has to be a list of one element,
        # or a single string of single path, otherwise quit
        if type(code) == list and len(code) == 1:
            entry_code = code[0]
        elif type(code) == str:
            entry_code = code
        else:
            logger.error(
                f"You have to specify an entry_code of your code: [{code}]")
            quit()

    # patch
    if patch_path != "":
        sp_patch = subprocess.Popen("patch -d/ -p0 < " + patch_path,
                                    env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
        sp_patch.wait()

    # if mode is "self_defined", user must specify ncpi, nins and bs
    if mode == "self_defined":
        if ncore_per_instance == -1 or ninstances == -1 or bench_batch_size == -1:
            logger.error(
                f"You have to specify ncore_per_instance, ninstances and bench_batch_size for \
                    self-defined benchmark mode.")
            quit()

    # numactl
    from . import numa_launcher

    from .utils.cpu_info import get_num_cpu_cores
    ncores = get_num_cpu_cores()

    # numactl setup for different modes
    if mode == "throughput":
        ncore_per_instance = ncores
        ninstances = 1
        bench_batch_size = 2 * ncores
    elif mode == "multi_instance":
        ncore_per_instance = 4
        ninstances = int(ncores / ncore_per_instance)
        bench_batch_size = 1
    elif mode == "latency":
        ncore_per_instance = 1
        ninstances = ncores
        bench_batch_size = 1
    elif mode == "self_defined":
        ncore_per_instance = ncore_per_instance
        ninstances = ninstances
        bench_batch_size = bench_batch_size

    # set cpu env variables
    if cpu_set_env:
        cmd_env = ''
        cmd_env += 'export LD_PRELOAD=${CONDA_PREFIX}/lib/libjemalloc.so'
        cmd_env += ' && '
        cmd_env += 'export LD_PRELOAD=${LD_PRELOAD}:${CONDA_PREFIX}/lib/libiomp5.so'
        cmd_env += ' && '
        cmd_env += 'export MALLOC_CONF="oversize_threshold:1,background_thread:true,metadata_thp:auto,\
            dirty_decay_ms:9000000000,muzzy_decay_ms:9000000000"'
        cmd_env += ' && '
        cmd_env += 'export KMP_AFFINITY="granularity=fine,compact,1,0"'
        cmd_env += ' && '
        cmd_env += 'export KMP_BLOCKTIME=1'
        cmd_env += ' && '
        cmd_env += 'export DNNL_PRIMITIVE_CACHE_CAPACITY=1024'
        cmd_env += ' && '
        cmd_env += 'export KMP_SETTINGS=1'

        sp_set_env = subprocess.Popen(
            cmd_env, env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
        sp_set_env.wait()

    # benchmark
    logger.info(f"Start benchmark on the code ...")

    bench_log_path = ws_path + "performance.log"
    os.remove(bench_log_path) if os.path.exists(bench_log_path) else 0

    entry_code_args = [entry_code_args]
    numa_launcher.exec_launcher(
        ncore_per_instance, ninstances, entry_code, entry_code_args, bench_log_path)

    # get performance (throughput and latency)
    bench_log = open(bench_log_path, "r").read().split('\n')
    IPS = 0
    MSPI = 0
    count_MSPI = 0
    P50 = 0
    count_P50 = 0
    P90 = 0
    count_P90 = 0
    P99 = 0
    count_P99 = 0
    for line in bench_log:
        if "Neural_Coder_Bench_IPS" in line:
            try:
                IPS += float(line[line.find(":")+3:])
            except ValueError as ve:
                pass
        if "Neural_Coder_Bench_MSPI" in line:
            try:
                MSPI += float(line[line.find(":")+3:])
                count_MSPI += 1
            except ValueError as ve:
                pass
        if "Neural_Coder_Bench_P50" in line:
            try:
                P50 += float(line[line.find(":")+3:])
                count_P50 += 1
            except ValueError as ve:
                pass
        if "Neural_Coder_Bench_P90" in line:
            try:
                P90 += float(line[line.find(":")+3:])
                count_P90 += 1
            except ValueError as ve:
                pass
        if "Neural_Coder_Bench_P99" in line:
            try:
                P99 += float(line[line.find(":")+3:])
                count_P99 += 1
            except ValueError as ve:
                pass

    FPS = round(IPS * bench_batch_size, 3)
    try:
        MSPI = round(MSPI / count_MSPI, 3)
    except:
        MSPI = 0
    try:
        P50 = round(P50 / count_P50, 3)
    except:
        MSPI = 0
    try:
        P90 = round(P90 / count_P90, 3)
    except:
        MSPI = 0
    try:
        P99 = round(P99 / count_P99, 3)
    except:
        MSPI = 0

    logger.info(f"Collected throughput on the code is: [{FPS}] (fps)")
    logger.info(f"Collected latency on the code is: [{MSPI}] (mspi)")
    logger.info(f"Collected latency_p50 on the code is: [{P50}] (mspi)")
    logger.info(f"Collected latency_p90 on the code is: [{P90}] (mspi)")
    logger.info(f"Collected latency_p99 on the code is: [{P99}] (mspi)")

    # unpatch
    if patch_path != "":
        sp_unpatch = subprocess.Popen(
            "patch -R -d/ -p0 < " + patch_path, env=os.environ, shell=True, stdout=subprocess.PIPE)  # nosec
        sp_unpatch.wait()

    return [FPS, MSPI, P50, P90, P99], mode, os.path.abspath(ws_path)


def superbench(
    code,
    entry_code="",
    entry_code_args="",
    sweep_objective="feature",
    bench_feature=[],
    mode="throughput",
    num_benchmark_iteration=30,
    logging_level="info",
    cpu_conversion=True,
    cpu_set_env=True,
    ncore_per_instance=-1,  # only for "self_defined" mode
    ninstances=-1,  # only for "self_defined" mode
    bench_batch_size=-1,  # only for "self_defined" mode
):

    # set up workspace
    ws_path = "neural_coder_workspace/" + \
        "superbench" + str(int(time.time())) + "/"
    os.makedirs(ws_path)

    # set up logging
    logging_var = "logging." + logging_level.upper()
    globals.logging_level = eval(logging_var)

    logger = logging.getLogger(ws_path)
    logger.setLevel(globals.logging_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s: - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

    fh = logging.FileHandler(ws_path+'superbench.log')
    fh.setLevel(globals.logging_level)
    fh.setFormatter(formatter)
    ch = logging.StreamHandler()
    ch.setLevel(globals.logging_level)
    ch.setFormatter(formatter)

    logger.addHandler(ch)
    logger.addHandler(fh)

    # print key inputs
    logger.info(f"Superbench started ...")
    logger.info(f"code: {code}")
    logger.info(f"mode: {mode}")
    logger.info(f"sweep_objective: {sweep_objective}")
    logger.info(f"num_benchmark_iteration: {num_benchmark_iteration}")

    # entry code
    if entry_code == "":
        # if not specify entry_code, then code has to be a list of one element,
        # or a single string of single path, otherwise quit
        if type(code) == list and len(code) == 1:
            entry_code = code[0]
        elif type(code) == str:
            entry_code = code
        else:
            logger.error(
                f"You have to specify an entry_code of your code: [{code}]")
            quit()

    if sweep_objective == "feature":
        list_FPS = []
        list_features = []
        list_mode = []
        list_ws_path = []
        result = []

        # features that is a "backend":
        backends = ["",
                    "pytorch_ipex_fp32",
                    "pytorch_ipex_bf16",
                    "pytorch_ipex_int8_static_quant",
                    "pytorch_ipex_int8_dynamic_quant",
                    "pytorch_inc_static_quant",
                    "pytorch_inc_dynamic_quant",
                    ]

        # features that can be standalone (either use alone or use with "backend"):
        standalones_pool = ["pytorch_channels_last",
                            "pytorch_mixed_precision_cpu",
                            "pytorch_jit_script",
                            "pytorch_jit_script_ofi",
                            "pytorch_torchdynamo_jit_script",
                            "pytorch_torchdynamo_jit_script_ofi",
                            "pytorch_torchdynamo_jit_trace",
                            "pytorch_torchdynamo_jit_trace_ofi",
                            ]

        if logging_level == "debug":
            # features that is a "backend":
            backends = ["",
                        "pytorch_ipex_fp32",
                        "pytorch_inc_static_quant",
                        ]

            # features that can be standalone (either use alone or use with "backend"):
            standalones_pool = ["pytorch_channels_last",
                                ]

        standalones = []
        standalones.append("")
        from itertools import combinations
        for num_items in range(len(standalones_pool)):
            list_comb = list(combinations(standalones_pool, num_items + 1))
            for item in list_comb:
                jit_feature_count = 0
                for i in list(item):
                    if "jit" in i:
                        jit_feature_count += 1
                if jit_feature_count <= 1:
                    standalones.append(list(item))  # only appends one JIT item

        for backend in backends:
            for standalone in standalones:
                features = []
                features.append(backend)
                features += standalone

                # exclude conflict features (like jit and jit_ofi)
                if "pytorch_ipex_fp32" in features and "pytorch_mixed_precision_cpu" in features:
                    continue
                if "pytorch_ipex_bf16" in features and "pytorch_mixed_precision_cpu" in features:
                    continue
                if "pytorch_ipex_int8_static_quant" in features and "pytorch_mixed_precision_cpu" in features:
                    continue
                if "pytorch_ipex_int8_dynamic_quant" in features and "pytorch_mixed_precision_cpu" in features:
                    continue
                if "pytorch_inc_static_quant" in features and "pytorch_mixed_precision_cpu" in features:
                    continue
                if "pytorch_inc_dynamic_quant" in features and "pytorch_mixed_precision_cpu" in features:
                    continue

                if cpu_conversion:
                    features.append("pytorch_cuda_to_cpu")

                if features[0] == "" and len(features) > 1:
                    features = features[1:]  # remove ""

                bench_performance, bench_mode, bench_ws_path = enable(
                    code=code,
                    entry_code=entry_code,
                    entry_code_args=entry_code_args,
                    features=features,
                    mode=mode,
                    run_bench=True,
                    num_benchmark_iteration=num_benchmark_iteration,
                    cpu_set_env=cpu_set_env,
                    ncore_per_instance=ncore_per_instance,
                    ninstances=ninstances,
                    bench_batch_size=bench_batch_size,
                )

                def remove_if_have(list, element):
                    if element in list:
                        list.remove(element)
                    return list

                features = remove_if_have(features, "pytorch_benchmark")
                features = remove_if_have(
                    features, "pytorch_change_batch_size")
                features = remove_if_have(features, "pytorch_cuda_to_cpu")

                logger.info(
                    f"Benchmark result of acceleration set [{features}] is [{bench_performance[0]}] (FPS)")

                d = {}  # initialize dict
                d["features"] = features
                d["FPS"] = bench_performance[0]
                d["mode"] = bench_mode
                d["workspace_path"] = bench_ws_path
                result.append(d)

                list_FPS.append(bench_performance[0])
                list_features.append(features)
                list_mode.append(bench_mode)
                list_ws_path.append(bench_ws_path)

        # print result
        logger.info(
            f"Superbench result of sweeping [{sweep_objective}] printed below with sorted FPS: ")
        print("{:<20} {:<20} {:<120}".format(
            'Numactl Mode', 'Performance (FPS)', 'Features Applied'))
        sort_index = sorted(range(len(list_FPS)),
                            key=lambda k: list_FPS[k], reverse=True)
        for i in sort_index:
            if list_FPS[i] != 0:
                print("{:<20} {:<20} {:<120}".format(
                    str(list_mode[i]), str(list_FPS[i]), str(list_features[i])))

        # for superbench report generation
        list_optimization_set_top3 = []
        list_performance_top3 = []
        count_top3 = 0
        for i in sort_index:
            if list_FPS[i] != 0:
                list_performance_top3.append(list_FPS[i])
                list_optimization_set_top3.append(list_features[i])
                count_top3 += 1
                if count_top3 == 3:
                    break

        original_model_performance = 0
        original_model_ranking = 0
        for i in sort_index:
            if list_FPS[i] != 0:
                original_model_ranking += 1
                if list_features[i] == []:
                    original_model_performance = list_FPS[i]
                    break

        return list_optimization_set_top3, \
            list_performance_top3, original_model_ranking, original_model_performance

    elif sweep_objective == "bench_config":
        result_ncpi = []
        result_nins = []
        result_bs = []
        result_regular_thp = []
        result_p50_thp = []
        result_p90_thp = []
        result_p99_thp = []
        if bench_feature == []:
            logger.error(
                f'You must specify a feature (optimization set) for benchmark when "sweep_objective" \
                    is "bench_config"')
            quit()
        else:
            from .utils.cpu_info import get_num_cpu_cores
            ncores = get_num_cpu_cores()
            list_ncpi = [1, 2, 4, 8]
            for i in [1, 2, 4, 8]:
                list_ncpi.append(int(ncores / i))
            list_ncpi = list(set(list_ncpi))
            list_ncpi.sort()
            logger.debug(f"list_ncpi = {list_ncpi}")

            for this_ncpi in list_ncpi:
                ncore_per_instance = this_ncpi
                ninstances = int(ncores / this_ncpi)
                list_bs = [1, 2, 4, 8, this_ncpi * 1, this_ncpi * 2, this_ncpi *
                           4, this_ncpi * 8, this_ncpi * 16, this_ncpi * 32, this_ncpi * 64]
                list_bs = list(set(list_bs))
                list_bs.sort()
                if logging_level == "debug":
                    list_bs = [list_bs[-5]]
                logger.debug(f"this_ncpi = {this_ncpi}")
                logger.debug(f"list_bs = {list_bs}")
                for this_bs in list_bs:
                    bench_batch_size = this_bs
                    try:
                        bench_performance, bench_mode, bench_ws_path = enable(
                            code=code,
                            entry_code=entry_code,
                            entry_code_args=entry_code_args,
                            features=bench_feature,
                            mode="self_defined",  # sweep bench_config, so mode set to "self_defined"
                            run_bench=True,
                            num_benchmark_iteration=num_benchmark_iteration,
                            cpu_set_env=cpu_set_env,
                            ncore_per_instance=ncore_per_instance,
                            ninstances=ninstances,
                            bench_batch_size=bench_batch_size,
                        )

                        socket_regular_thp = bench_performance[0]
                        socket_p50_thp = round(
                            1000 / bench_performance[2] * ninstances * bench_batch_size, 3)
                        socket_p90_thp = round(
                            1000 / bench_performance[3] * ninstances * bench_batch_size, 3)
                        socket_p99_thp = round(
                            1000 / bench_performance[4] * ninstances * bench_batch_size, 3)

                        result_ncpi.append(ncore_per_instance)
                        result_nins.append(ninstances)
                        result_bs.append(bench_batch_size)
                        result_regular_thp.append(socket_regular_thp)
                        result_p50_thp.append(socket_p50_thp)
                        result_p90_thp.append(socket_p90_thp)
                        result_p99_thp.append(socket_p99_thp)

                        logger.info(
                            f"ncpi: {ncore_per_instance}, nins: {ninstances}, bs: {bench_batch_size}, \
                                regular_thp: {socket_regular_thp}, p50_thp: {socket_p50_thp}, \
                                    p90_thp: {socket_p90_thp}, p99_thp: {socket_p99_thp}")

                    except:
                        logger.warning(
                            f"ncpi: {ncore_per_instance}, nins: {ninstances}, bs: {bench_batch_size}, Benchmark \
                                failed. It might be due to HW limitation such as CPU load limit.")
                        continue

        # print result
        for item in [result_regular_thp, result_p50_thp, result_p90_thp, result_p99_thp]:
            if item is result_regular_thp:
                display_item_name = "Throughput"
            elif item is result_p50_thp:
                display_item_name = "Throughput based on P50-Latency"
            elif item is result_p90_thp:
                display_item_name = "Throughput based on P90-Latency"
            elif item is result_p99_thp:
                display_item_name = "Throughput based on P99-Latency"

            print("{:<30} {:<30} {:<30} {:<30}".format(
                'Num Cores Per Instance', 'Num of Instances', 'Batch Size', display_item_name))
            sort_index = sorted(
                range(len(item)), key=lambda k: item[k], reverse=True)
            for i in sort_index:
                print("{:<30} {:<30} {:<30} {:<30}".format(str(result_ncpi[i]), str(
                    result_nins[i]), str(result_bs[i]), str(item[i])))

        list_config_best_ncpi = []
        list_config_best_nins = []
        list_config_best_bs = []
        list_config_best_performance = []
        for item in [result_regular_thp, result_p50_thp, result_p90_thp, result_p99_thp]:
            sort_index = sorted(
                range(len(item)), key=lambda k: item[k], reverse=True)
            for i in sort_index:
                list_config_best_ncpi.append(result_ncpi[i])
                list_config_best_nins.append(result_nins[i])
                list_config_best_bs.append(result_bs[i])
                list_config_best_performance.append(item[i])
                break  # only fetch the top result

        return list_config_best_ncpi, list_config_best_nins, list_config_best_bs, list_config_best_performance



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .. import globals

# [insert] some code lines into file


def execute_insert_transformation(list_transformed_code):
    for file_path in globals.list_trans_insert_modified_file:
        trans_location_idxs = globals.list_trans_insert_location_idxs[globals.list_trans_insert_modified_file.index(
            file_path)]
        trans_number_insert_lines = \
            globals.list_trans_insert_number_insert_lines[globals.list_trans_insert_modified_file.index(
                file_path)]
        trans_lines_to_insert = \
            globals.list_trans_insert_lines_to_insert[globals.list_trans_insert_modified_file.index(
                file_path)]

        file_path_idx = globals.list_code_path.index(file_path)
        lines_transformed = list_transformed_code[file_path_idx].split('\n')

        # this part is for "insert" kind of transformation only (math)
        t = [0]
        u = 0
        for n in trans_number_insert_lines:
            u = u + n
            t.append(u)
        t = t[:-1]

        trans_location_idxs = [sum(i) for i in zip(trans_location_idxs, t)]

        for idx in trans_location_idxs:  # actual transformation (insertion)
            additions = trans_lines_to_insert[trans_location_idxs.index(
                idx)].split("\n")
            additions = additions[::-1]  # reverse
            for i in range(len(additions)):
                lines_transformed.insert(idx, additions[i])

        # transfer lines_transformed to code format ("\n" save write)
        code_transformed = "".join([i + "\n" for i in lines_transformed])[0:-1]

        list_transformed_code[file_path_idx] = code_transformed

    return list_transformed_code

# [indenting] some code lines with " " into file


def execute_indenting_transformation(list_transformed_code):
    for file_path in globals.list_trans_indenting_modified_file:
        trans_location_idxs = \
            globals.list_trans_indenting_location_idxs[globals.list_trans_indenting_modified_file.index(
                file_path)]
        trans_indenting_level = \
            globals.list_trans_indenting_level[globals.list_trans_indenting_modified_file.index(
                file_path)]

        file_path_idx = globals.list_code_path.index(file_path)
        lines_transformed = list_transformed_code[file_path_idx].split('\n')

        for idx in trans_location_idxs:  # actual transformation (indenting)
            this_indenting_level = trans_indenting_level[trans_location_idxs.index(
                idx)]
            lines_transformed[idx] = " " * 4 * \
                this_indenting_level + lines_transformed[idx]

        # transfer lines_transformed to code format ("\n" save write)
        code_transformed = "".join([i + "\n" for i in lines_transformed])[0:-1]

        list_transformed_code[file_path_idx] = code_transformed

    return list_transformed_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
from schema import Schema, And, Use, Optional, Or, Hook
from ..adaptor import FRAMEWORKS
from ..strategy import STRATEGIES
from ..objective import OBJECTIVES
from ..pruners import PRUNERS
from ..utils import logger
from ..version import __version__
import re
import copy
import itertools
from collections import OrderedDict
from .dotdict import DotDict
import os, datetime

def constructor_register(cls):
    yaml_key = "!{}".format(cls.__name__)

    def constructor(loader, node):
        instance = cls.__new__(cls)
        yield instance

        state = loader.construct_mapping(node, deep=True)
        instance.__init__(**state)

    yaml.add_constructor(
        yaml_key,
        constructor,
        yaml.SafeLoader,
    )
    return cls

@constructor_register
class Pruner():
    def __init__(self, start_epoch=None, end_epoch=None, initial_sparsity=None,
                 target_sparsity=None, update_frequency=1, prune_type='basic_magnitude',
                 method='per_tensor', names=[], parameters=None):
        self.start_epoch = start_epoch
        self.end_epoch = end_epoch
        self.update_frequency = update_frequency
        self.target_sparsity = target_sparsity
        self.initial_sparsity = initial_sparsity
        self.update_frequency = update_frequency
        assert prune_type.replace('_', '') in [i.lower() for i in PRUNERS], \
                                         'now only support {}'.format(PRUNERS.keys())
        self.prune_type = prune_type
        self.method = method
        self.names= names
        self.parameters = parameters

# Schema library has different loading sequence priorities for different
# value types.
# To make sure the fields under dataloader.transform field of yaml file
# get loaded with written sequence, this workaround is used to convert
# None to {} in yaml load().
yaml.SafeLoader.add_constructor('tag:yaml.org,2002:null', lambda loader, node: {})
# Add python tuple support because best_configure.yaml may contain tuple
yaml.SafeLoader.add_constructor('tag:yaml.org,2002:python/tuple',
                                lambda loader, node: tuple(loader.construct_sequence(node)))

def _valid_accuracy_field(key, scope, error):
    assert bool(
        'relative' in scope['accuracy_criterion']) != bool(
        'absolute' in scope['accuracy_criterion'])

def _valid_prune_epoch(key, scope, error):
    if "start_epoch" in scope[key] and "end_epoch" in scope[key]:
        assert scope[key]["start_epoch"] <= scope[key]["end_epoch"]

def _valid_prune_sparsity(key, scope, error):
    if "initial_sparsity" in scope[key] and "target_sparsity" in scope[key]:
        assert scope[key]["initial_sparsity"] <= scope[key]["target_sparsity"]
    elif "initial_sparsity" in scope[key]:
        assert scope[key]["initial_sparsity"] >= 0
    elif "target_sparsity" in scope[key]:
        assert scope[key]["target_sparsity"] < 1

def _valid_multi_objectives(key, scope, error):
    if 'weight' in scope[key] and scope[key]['weight'] is not None:
        assert len(scope[key]['objective']) == len(scope[key]['weight'])

def _valid_multi_metrics(key, scope, error):
    if 'metric' in scope and 'multi_metrics' in scope:
        assert False

def _valid_metric_length(key, scope, error):
    metrics = [i for i in scope[key] if i != 'weight' and i != 'higher_is_better']
    if 'weight' in scope[key] and scope[key]['weight'] is not None:
        assert len(input_to_list_float(scope[key]['weight'])) == len(metrics)
    if 'higher_is_better' in scope[key] and scope[key]['higher_is_better'] is not None:
        assert len(input_to_list_bool(scope[key]['higher_is_better'])) == len(metrics)

# used for '123.68 116.78 103.94' style to float list
def input_to_list_float(data):
    if isinstance(data, str):
        return [float(s.strip()) for s in data.split()]

    if isinstance(data, float):
        return [data]

    assert isinstance(data, list)
    return [float(d) for d in data]

def input_to_list_bool(data):
    if isinstance(data, str):
        if ',' in data:
            return [s.strip() == 'True' for s in data.split(',')]
        else:
            return [s.strip() == 'True' for s in data.split()]

    if isinstance(data, bool):
        return [data]

    assert isinstance(data, list) and all([isinstance(i, bool) for i in data])
    return data

def input_int_to_float(data):
    if isinstance(data, str):
        # used for '123.68, 116.78, 103.94' style
        if ',' in data:
            data = data.split(',')
        # used for '123.68 116.78 103.94' style
        else:
            data = data.split()

        if len(data) == 1:
            return float(data[0].strip())
        else:
            return [float(s.strip()) for s in data]
    elif isinstance(data, list):
        return [float(s) for s in data]
    elif isinstance(data, int):
        return float(data)

def input_to_list_int(data):
    if isinstance(data, str):
        return [int(s.strip()) for s in data.split(',')]

    if isinstance(data, int):
        return [data]

    assert isinstance(data, list)
    return [int(d) for d in data]

def input_to_list(data):
    if isinstance(data, str):
        if ',' in data:
            return [s.strip() for s in data.split(',')]

        return [s.strip() for s in data.split()]

    if isinstance(data, int):
        return [data]

    assert isinstance(data, list)
    return data

def list_to_tuple(data):
    if isinstance(data, str):
        return tuple([int(s.strip()) for s in data.split(',')])

    elif isinstance(data, list):
        if isinstance(data[0], list):
            result = []
            for item in data:
                result.append(tuple([int(s) for s in item]))
            return result
        else:
            return tuple([int(s) for s in data])

def percent_to_float(data):
    if isinstance(data, str) and re.match(r'-?\d+(\.\d+)?%', data):
        data = float(data.strip('%')) / 100
    if isinstance(data, int):
        data = float(data)
    else:
        assert isinstance(data, float), 'This field should be float, int or percent string'
    return data

ops_schema = Schema({
    Optional('weight', default=None): {
        Optional('granularity'): And(
            list,
            lambda s: all(i in ['per_channel', 'per_tensor'] for i in s)),
        Optional('scheme'): And(
            list,
            # asym_float are only for PyTorch framework
            lambda s: all(i in ['asym', 'sym', 'asym_float'] for i in s)),
        Optional('dtype'): And(
            list,
            lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'fp16'] for i in s)),
        Optional('algorithm'): And(
            list,
            lambda s: all(i in ['minmax'] for i in s)),
        Optional('bit'):  And(
            Or(float, list),
            Use(input_to_list_float),
            lambda s: all(0.0 < i <= 7.0 for i in s))
    },
    Optional('activation', default=None): {
        Optional('granularity'): And(
            list,
            lambda s: all(i in ['per_channel', 'per_tensor'] for i in s)),
        Optional('scheme'): And(
            list,
            lambda s: all(i in ['asym', 'sym'] for i in s)),
        Optional('dtype'): And(
            list,
            lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'fp16'] for i in s)),
        # compute_dtypeis only for PyTorch framework
        Optional('compute_dtype', default=['uint8']): And(
            list,
            lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'None'] for i in s)),
        # placeholder are only for PyTorch framework
        Optional('algorithm'): And(
            list,
            lambda s: all(i in ['minmax', 'kl', 'placeholder'] for i in s))
    }
})

graph_optimization_schema = Schema({

    Optional('precisions', default={'precisions': ['fp32']}): And(
        Or(str, list),
        Use(input_to_list),
        lambda s: all(i in [ 'fp32', 'bf16'] for i in s)),

    Optional('op_wise', default={'weight': {}, 'activation': {}}): {
        Optional('weight', default=None): {
            Optional('dtype', default=None): And(
                Or(str, list),
                Use(input_to_list),
                lambda s: all(i in ['fp32', 'bf16', 'fp16'] for i in s)),
        },
        Optional('activation', default=None): {
            Optional('dtype', default=None): And(
                Or(str, list),
                Use(input_to_list),
                lambda s: all(i in ['fp32', 'bf16', 'fp16'] for i in s)),
            }
    }
})

mixed_precision_schema = Schema({

    Optional('precisions', default={'precisions': ['fp32']}): And(
        Or(str, list),
        Use(input_to_list),
        lambda s: all(i in [ 'fp32', 'bf16', 'fp16'] for i in s)),

    Optional('op_wise', default={'weight': {}, 'activation': {}}): {
        Optional('weight', default=None): {
            Optional('dtype', default=None): And(
                Or(str, list),
                Use(input_to_list),
                lambda s: all(i in ['fp32', 'bf16', 'fp16'] for i in s)),
        },
        Optional('activation', default=None): {
            Optional('dtype', default=None): And(
                Or(str, list),
                Use(input_to_list),
                lambda s: all(i in ['fp32', 'bf16', 'fp16'] for i in s)),
            }
    }
})

model_conversion_schema = Schema({
    'source': And(str, lambda s: s.lower() == 'qat'),
    'destination': And(str, lambda s: s.lower() == 'default')
})

filter_schema = Schema({
    Optional('LabelBalance'): {
        'size': And(int, lambda s: s > 0)
    },
})

transform_schema = Schema({
    Optional('ResizeWithRatio'):{
        Optional('min_dim'): int,
        Optional('max_dim'): int,
        Optional('padding'): bool,
        Optional('constant_value'): int
    },
    Optional('CropToBoundingBox'): {
        'offset_height': int,
        'offset_width': int,
        'target_height': int,
        'target_width': int
    },
    Optional('Cast'): {
        Optional('dtype'): str
    },
    Optional('RandomResizedCrop'): {
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0)),
        Optional('scale'): And(list, lambda s: all(isinstance(i, float) for i in s)),
        Optional('ratio'): And(list, lambda s: all(isinstance(i, float) for i in s)),
        Optional('interpolation'): And(
            str,
            lambda s: s in ['nearest', 'bilinear', 'bicubic']),
    },
    Optional('AlignImageChannel'): {
        Optional('dim'): int
    },
    Optional('ToNDArray'): Or({}, None),
    Optional('CropResize'): {
        'x': int,
        'y': int,
        'width': int,
        'height': int,
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0)),
        Optional('interpolation'): And(
            str,
            lambda s: s in ['nearest', 'bilinear', 'bicubic']),
    },
    Optional('RandomHorizontalFlip'): Or({}, None),
    Optional('RandomVerticalFlip'): Or({}, None),
    Optional('ToTensor'): Or({}, None),
    Optional('ToPILImage'): Or({}, None),
    Optional('Normalize'): {
        Optional('mean'): And(list, lambda s: all(isinstance(i, float) for i in s)),
        Optional('std'): And(list, lambda s: all(isinstance(i, float) for i in s)),
        Optional('rescale'): list
    },
    Optional('Resize'): {
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0)),
        Optional('interpolation'): And(
            str,
            lambda s: s in ['nearest', 'bilinear', 'bicubic']),
    },
    Optional('RandomCrop'): {
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0))
    },
    Optional('Rescale'): Or({}, None),
    Optional('CenterCrop'): {
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0))
    },
    Optional('PaddedCenterCrop'): {
        'size': Or(And(list, lambda s: all(isinstance(i, int) for i in s)),
                    And(int, lambda s: s > 0)),
        Optional('crop_padding'): And(int, lambda s: s > 0),
    },
    Optional('ToArray'): Or({}, None),
    Optional('QuantizedInput'): {
        Optional('dtype', default='int8'): And(str, lambda s: s in ['int8', 'uint8']),
        Optional('scale'): And(float, lambda s: s > 0),
    },
    Optional('Transpose'): {
        'perm': And(list, lambda s: all(isinstance(i, int) for i in s)),
    },
    # THIS API IS TO BE DEPRECATED!
    Optional('ParseDecodeImagenet'): Or({}, None),
    Optional('ParseDecodeCoco'): Or({}, None),
    Optional('ParseDecodeVoc'): Or({}, None),
    Optional('BilinearImagenet'): {
        'height': And(int, lambda s: s > 0),
        'width': And(int, lambda s: s > 0),
        Optional('central_fraction'): float,
        Optional('mean_value'): And(Or(str, list), Use(input_to_list_float)),
        Optional('scale'): float,
    },
    Optional('ResizeCropImagenet'): {
        'height': And(int, lambda s: s > 0),
        'width': And(int, lambda s: s > 0),
        Optional('random_crop'): bool,
        Optional('slice_crop'): bool,
        Optional('resize_side'): And(int, lambda s: s > 0),
        Optional('resize_method', default='bilinear'): \
            And(str, lambda s: s in ['bilinear', 'lanczos3', 'lanczos5',
                                     'bicubic', 'gaussian', 'nearest',
                                     'area', 'mitchellcubic']),
        Optional('random_flip_left_right'): bool,
        Optional('data_format', default='channels_last'): \
            And(str, lambda s: s in ['channels_first', 'channels_last']),
        Optional('subpixels', default='RGB'): \
            And(str, lambda s: s in ['BGR', 'RGB']),
        Optional('mean_value'): And(Or(str, list), Use(input_to_list_float)),
        Optional('scale'): float,
    },
    Optional('ResizeWithAspectRatio'):{
        'height': And(int, lambda s: s > 0),
        'width': And(int, lambda s: s > 0),
    },
    Optional('ParseDecodeImagenet'): Or({}, None),
    Optional('ToArray'): Or({}, None),
    Optional('QuantizedInput'): {
        Optional('dtype', default='int8'): And(str, lambda s: s in ['int8', 'uint8']),
        Optional('scale'): And(float, lambda s: s > 0),
    },
    Optional('Transpose'): {
        'perm': And(list, lambda s: all(isinstance(i, int) for i in s)),
    },
})

postprocess_schema = Schema({
    Optional('LabelShift'):  int,
    Optional('Collect'): {
        'length': int
    },
    Optional('SquadV1'): {
        'label_file': str,
        'vocab_file': str,
        Optional('do_lower_case', default='True'): bool,
        Optional('max_seq_length', default=384): int,
    },
    Optional('SquadV1ModelZoo'): {
        'label_file': str,
        'vocab_file': str,
        Optional('do_lower_case', default='True'): bool,
        Optional('max_seq_length', default=384): int,
    },
})

dataset_schema = Schema({
    Optional('CIFAR10'): {
        'root': str,
        Optional('train'): bool,
        Optional('download'): bool,
    },
    Optional('CIFAR100'): {
        'root': str,
        Optional('train'): bool,
        Optional('download'): bool,
    },
    Optional('MNIST'): {
        'root': str,
        Optional('train'): bool,
        Optional('download'): bool,
    },
    Optional('FashionMNIST'): {
        'root': str,
        Optional('train'): bool,
        Optional('download'): bool,
    },
    Optional('ImageFolder'): {
        'root': str,
    },
    Optional('TFRecordDataset'): {
        'root': str,
    },
    Optional('ImageRecord'): {
        'root': str,
    },
    Optional('dummy_v2'): {
        'input_shape': And(Or(str, list), Use(list_to_tuple)),
        Optional('label_shape', default=[1]): And(Or(str, list), Use(list_to_tuple)),
        Optional('low'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('high'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('dtype'): And(Or(str, list), Use(input_to_list)),
    },
    Optional('sparse_dummy_v2'): {
        'dense_shape': And(Or(str, list), Use(list_to_tuple)),
        Optional('label_shape', default=[1]): And(Or(str, list), Use(list_to_tuple)),
        Optional('sparse_ratio'): Or(
            float,
            And(list, Use(input_int_to_float)),
            And(int, Use(input_int_to_float))),
        Optional('low'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('high'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('dtype'): And(Or(str, list), Use(input_to_list)),
    },
 
    Optional('dummy'): {
        'shape': And(Or(str, list), Use(list_to_tuple)),
        Optional('low'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('high'): Or(
            float,
            And(int, Use(input_int_to_float)),
            And(list, Use(input_int_to_float)),
            And(str, Use(input_int_to_float))),
        Optional('dtype'): And(Or(str, list), Use(input_to_list)),
        Optional('label'): bool,
    },
    Optional('bert'): {
        'root': str,
        'label_file': str,
        Optional('task'): And(str, lambda s: s in ["classifier", "squad"]),
        Optional('model_type'): And(str, lambda s: s in ['bert', 'xlnet', 'xlm']),
    },
    Optional('mzbert'): {
        'root': str,
        'label_file': str,
        Optional('task'): And(str, lambda s: s in ["classifier", "squad"]),
        Optional('model_type'): And(str, lambda s: s in ['bert', 'xlnet', 'xlm']),
    },
    Optional('VOCRecord'): {
        'root': str,
    },
    Optional('COCORecord'): {
        'root': str,
        Optional('num_cores'): int,
    },
    Optional('COCORaw'): {
        'root': str,
        Optional('img_dir'): str,
        Optional('anno_dir'): str,
        Optional('num_cores'): int,
    },
    Optional('COCONpy'): {
        'root': str,
        'npy_dir': str,
        Optional('anno_dir'): str,
        Optional('num_cores'): int,
    },
    Optional('ImagenetRaw'): {
        'data_path': str,
        Optional('image_list'): str,
    },
    Optional('style_transfer'): {
        'content_folder': str,
        'style_folder': str,
        Optional('crop_ratio'): float,
        Optional('resize_shape'): And(Or(str, list), Use(input_to_list_int)),
        Optional('image_format'): str,
    },
    Optional('GLUE'): {
        'data_dir': str,
        'model_name_or_path': str,
        Optional('max_seq_length'): int,
        Optional('do_lower_case'): bool,
        Optional('task'): str,
        Optional('model_type'): str,
        Optional('dynamic_length'): bool,
        Optional('evaluate'): bool
    },
    # TO BE DEPRECATED!
    Optional('Imagenet'): {
        'root': str,
    },
})

dataloader_schema = Schema({
    Optional('last_batch', default='rollover'): And(str, lambda s: s in ['rollover', 'discard']),
    Optional('batch_size', default=None): And(int, lambda s: s > 0),
    'dataset': dataset_schema,
    Optional('filter'): filter_schema,
    Optional('transform'): transform_schema,
    Optional('shuffle', default = False): And(bool, lambda s: s in [True, False]),
    Optional('distributed', default = False): And(bool, lambda s: s in [True, False]),
})

configs_schema = Schema({
    Optional('cores_per_instance'): And(int, lambda s: s > 0),
    Optional('num_of_instance', default=1): And(int, lambda s: s > 0),
    Optional('inter_num_of_threads'): And(int, lambda s: s > 0),
    Optional('intra_num_of_threads'): And(int, lambda s: s > 0),
    Optional('kmp_blocktime'): And(int, lambda s: s >= 0),
    Optional('kmp_affinity', default='granularity=fine,verbose,compact,1,0'): str,
})

optimizer_schema = Schema({
    Optional('SGD'): {
        'learning_rate': Use(float),
        Optional('momentum'): Use(float),
        Optional('nesterov'): bool,
        Optional('weight_decay'): Use(float)
    },
    Optional('AdamW'): {
        'weight_decay': Use(float),
        Optional('learning_rate', default=0.001): Use(float),
        Optional('beta_1', default=0.9): Use(float),
        Optional('beta_2', default=0.999): Use(float),
        Optional('epsilon', default=1e-07): Use(float),
        Optional('amsgrad', default=False): bool
    },
    Optional('Adam'): {
        Optional('learning_rate', default=0.001): Use(float),
        Optional('beta_1', default=0.9): Use(float),
        Optional('beta_2', default=0.999): Use(float),
        Optional('epsilon', default=1e-07): Use(float),
        Optional('amsgrad', default=False): bool
    }, 
})

criterion_schema = Schema({
    Optional('CrossEntropyLoss'): {
        Optional('reduction', default='mean'): \
            And(str, lambda s: s in ['none', 'sum', 'mean', 'auto', 'sum_over_batch_size']),
        Optional('from_logits', default=False):
            And(bool, lambda s: s in [True, False]),
    },
    Optional('SparseCategoricalCrossentropy'): {
        Optional('reduction', default='mean'): \
            And(str, lambda s: s in ['none', 'sum', 'mean', 'auto', 'sum_over_batch_size']),
        Optional('from_logits', default=False):
            And(bool, lambda s: s in [True, False]),
    },
    Optional('KnowledgeDistillationLoss'): {
        Optional('temperature'): And(float, lambda s: s > 0),
        Optional('loss_types'): And(list, lambda s: all(i in ['CE', 'KL'] for i in s)),
        Optional('loss_weights'): And(list, lambda s: all(i >= 0 for i in s) and sum(s) == 1.0),
    },
    Optional('IntermediateLayersKnowledgeDistillationLoss'): {
        'layer_mappings':
            And(Or(tuple, list), lambda s: all(len(i) in [2, 4] for i in s)),
        Optional('loss_types'):
            And(Or(tuple, list), lambda s: all(i in ['MSE', 'KL', 'L1'] for i in s)),
        Optional('loss_weights'):
            And(Or(tuple, list), lambda s: all(i >= 0 for i in s) and sum(s) == 1.0),
        Optional('add_origin_loss'): bool,
    }
})

train_schema = Schema({
    'optimizer': optimizer_schema,
    'criterion': criterion_schema,
    Optional('dataloader'): dataloader_schema,
    Optional('epoch', default=1): int,
    Optional('start_epoch', default=0): int,
    Optional('end_epoch'): int,
    Optional('iteration'): int,
    Optional('frequency'): int,
    Optional('execution_mode', default='eager'): And(str, lambda s: s in ['eager', 'graph']),
    Optional('postprocess'): {
        Optional('transform'): postprocess_schema
    },
    # TODO reserve for multinode training support
    Optional('hostfile'): str
})

weight_compression_schema = Schema({
    Optional('initial_sparsity', default=0): And(float, lambda s: s < 1.0 and s >= 0.0),
    Optional('target_sparsity', default=0.97): float,
    Optional('start_epoch', default=0): int,
    Optional('end_epoch', default=4): int,
    Optional('pruners'): And(list, \
                               lambda s: all(isinstance(i, Pruner) for i in s))
})

approach_schema = Schema({
    Hook('weight_compression', handler=_valid_prune_sparsity): object,
    Optional('weight_compression'): weight_compression_schema,
})

default_workspace = './nc_workspace/{}/'.format(
                                           datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))

COCOmAP_input_order_schema = Schema({
    Optional('num_detections'): int,
    'boxes': int,
    'scores': int,
    'classes': int
})

schema = Schema({
    'model': {
        'name': str,
        'framework': And(str, lambda s: s in list(FRAMEWORKS.keys()) + ['NA']),
        Optional('inputs', default=[]): And(Or(str, list), Use(input_to_list)),
        Optional('outputs', default=[]): And(Or(str, list), Use(input_to_list)),
    },
    Optional('version', default=float(__version__.split('.')[0])): And(
                                          Or(float,
                                             And(int, Use(input_int_to_float)),
                                             And(str, Use(input_int_to_float))),
                                          lambda s: s == float(__version__.split('.')[0])),
    Optional('device', default='cpu'): And(str, lambda s: s in ['cpu', 'gpu']),
    Optional('quantization', default={'approach': 'post_training_static_quant', \
                                      'calibration': {'sampling_size': [100]}, \
                                      'recipes': {'scale_propagation_max_pooling': True,
                                                      'scale_propagation_concat': True,
                                                      'first_conv_or_matmul_quantization': True},
                                      'model_wise': {'weight': {'bit': [7.0]},
                                                     'activation': {}},
                                      }): {
        Optional('approach', default='post_training_static_quant'): And(
            str,
            # TODO check if framework support dynamic quantize
            # Now only onnruntime and pytorch supoort
            lambda s: s in ['post_training_static_quant',
                            'post_training_dynamic_quant',
                            'quant_aware_training']),
        Optional('train', default=None): train_schema,
        Optional('advance', default=None): {
            Optional('bias_correction'): And(str, lambda s: s in ['weight_empirical']),
        },
        Optional('calibration', default={'sampling_size': [100]}): {
            Optional('sampling_size', default=[100]): And(Or(str, int, list), Use(input_to_list)),
            Optional('dataloader', default=None): dataloader_schema
        },
        Optional('recipes', default={'scale_propagation_max_pooling': True,
                                         'scale_propagation_concat': True,
                                         'first_conv_or_matmul_quantization': True}): {
            Optional('scale_propagation_max_pooling', default=True):
                    And(bool, lambda s: s in [True, False]),
            Optional('scale_propagation_concat', default=True):
                    And(bool, lambda s: s in [True, False]),
            Optional('first_conv_or_matmul_quantization', default=True):
                    And(bool, lambda s: s in [True, False]),
            Optional('fast_bias_correction', default=False):
                    And(bool, lambda s: s in [True, False]),
            Optional('weight_correction', default=False):
                    And(bool, lambda s: s in [True, False]),
        },
        Optional('model_wise', default={'weight': {'bit': [7.0]}, 'activation': {}}): {
            Optional('weight', default= {'bit': [7.0]}): {
                Optional('granularity', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['per_channel', 'per_tensor'] for i in s)),
                # asym_float are only for PyTorch framework
                Optional('scheme', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['asym', 'sym', 'asym_float'] for i in s)),
                Optional('dtype', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'fp16'] for i in s)),
                Optional('algorithm', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['minmax'] for i in s)),
                Optional('bit', default=[7.0]):  And(
                    Or(float, list),
                    Use(input_to_list_float),
                    lambda s: all(0.0 < i <= 7.0 for i in s))

            },
            Optional('activation', default=None): {
                Optional('granularity', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['per_channel', 'per_tensor'] for i in s)),
                Optional('scheme', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['asym', 'sym'] for i in s)),
                Optional('dtype', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'fp16'] for i in s)),
                # compute_dtypeis only for PyTorch framework
                Optional('compute_dtype', default=['uint8']): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['int8', 'uint8', 'fp32', 'bf16', 'None'] for i in s)),
                # placeholder are only for PyTorch framework
                Optional('algorithm', default=None): And(
                    Or(str, list),
                    Use(input_to_list),
                    lambda s: all(i in ['minmax', 'kl', 'placeholder'] for i in s)),
            }
        },
        Optional('optype_wise', default=None): {
            str: ops_schema
        },
        Optional('op_wise', default=None): {
            str: ops_schema
        },
    },

    Optional('graph_optimization'): graph_optimization_schema,
    Optional('mixed_precision'): mixed_precision_schema,

    Optional('model_conversion'): model_conversion_schema,

    Optional('tuning', default={
        'strategy': {'name': 'basic'},
        'accuracy_criterion': {'relative': 0.01, 'higher_is_better': True},
        'objective': 'performance',
        'exit_policy': {'timeout': 0, 'max_trials': 100, 'performance_only': False},
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace},
        'diagnosis': False,
        }): {
        Optional('strategy', default={'name': 'basic'}): {
            'name': And(str, lambda s: s in STRATEGIES), Optional('sigopt_api_token'): str,
            Optional('sigopt_project_id'): str,
            Optional('sigopt_experiment_name', default='nc-tune'): str,
            Optional('accuracy_weight', default=1.0): float,
            Optional('latency_weight', default=1.0): float
        } ,
        Hook('accuracy_criterion', handler=_valid_accuracy_field): object,
        Optional('accuracy_criterion', default={'relative': 0.01}): {
            Optional('relative'): And(Or(str, float), Use(percent_to_float)),
            Optional('absolute'): And(Or(str, int, float), Use(percent_to_float)),
            Optional('higher_is_better', default=True): bool,
        },
        Optional('objective', default='performance'): And(str, lambda s: s in OBJECTIVES),
        Hook('multi_objectives', handler=_valid_multi_objectives): object,
        Optional('multi_objectives'):{ 
            Optional('objective'): And(
                Or(str, list), Use(input_to_list), lambda s: all(i in OBJECTIVES for i in s)),
            Optional('weight'): And(Or(str, list), Use(input_to_list_float)),
            Optional('higher_is_better'): And(
                Or(str, bool, list), Use(input_to_list_bool)),
        },
        Optional('exit_policy', default={'timeout': 0,
                                         'max_trials': 100,
                                         'performance_only': False}): {
            Optional('timeout', default=0): int,
            Optional('max_trials', default=100): int,
            Optional('performance_only', default=False): bool,
        },
        Optional('random_seed', default=1978): int,
        Optional('tensorboard', default=False): And(bool, lambda s: s in [True, False]),
        Optional('workspace', default={'path': default_workspace}): {
            Optional('path', default=None): str,
            Optional('resume'): str
        },
        Optional('diagnosis', default = {
            'diagnosis_after_tuning': False,
            'op_list': [],
            'iteration_list': [1],
            'inspect_type': 'activation',
            'save_to_disk': True,
            'save_path': './nc_workspace/inspect_saved/',
        }):{
            Optional('diagnosis_after_tuning', default=False): And(bool, lambda s: s in [True, False]),
            Optional('op_list', default=[]): And(Or(str, list), Use(input_to_list)),
            Optional('iteration_list', default=[1]): And(Or(int, list), Use(input_to_list_int)),
            Optional('inspect_type', default='all'): And(str, lambda s : s in ['all', 'activation', 'weight']),
            Optional('save_to_disk', default=True): And(bool, lambda s: s in [True, False]),
            Optional('save_path', default='./nc_workspace/inspect_saved/'): str,
        },
    },
    Optional('evaluation'): {
        Hook('accuracy', handler=_valid_multi_metrics): object,
        Optional('accuracy'): {
            Hook('multi_metrics', handler=_valid_metric_length): object,
            Optional('multi_metrics', default=None): {
                Optional('weight'): And(Or(str, list), Use(input_to_list_float)),
                Optional('higher_is_better'): And(
                    Or(str, list), Use(input_to_list_bool)),
                Optional('topk'): And(int, lambda s: s in [1, 5]),
                Optional('mAP'): {
                    Optional('anno_path'): str,
                    Optional('iou_thrs', default=0.5):
                            Or(And(str, lambda s: s in ['0.5:0.05:0.95']),
                               And(float, lambda s: s <= 1.0 and s >= 0.0)),
                    Optional('map_points', default=0): And(int, lambda s: s in [0, 11, 101])
                },
                Optional('COCOmAP'): {
                    Optional('anno_path'): str,
                    Optional('map_key', default='DetectionBoxes_Precision/mAP'): str
                },
                Optional('COCOmAPv2'): {
                    Optional('anno_path'): str,
                    Optional('map_key', default='DetectionBoxes_Precision/mAP'): str,
                    Optional('output_index_mapping', default={'num_detections': -1, 
                                                      'boxes': 0, 
                                                      'scores': 1, 
                                                      'classes': 2}): COCOmAP_input_order_schema
                },
                Optional('VOCmAP'): {
                    Optional('anno_path'): str
                },
                Optional('SquadF1'): Or({}, None),
                Optional('MSE'): {
                    Optional('compare_label'): bool
                },
                Optional('RMSE'): {
                    Optional('compare_label'): bool
                },
                Optional('MAE'): {
                    Optional('compare_label'): bool
                },
                Optional('Accuracy'): Or({}, None),
                Optional('Loss'): Or({}, None),
                Optional('BLEU'): Or({}, None),
                Optional('SquadF1'): Or({}, None),
                Optional('F1'): Or({}, None),
                Optional('mIOU'): {
                    Optional('num_classes'): int
                },
                Optional('GLUE'): {
                    Optional('task'): str
                },
                Optional('ROC'): {
                    Optional('task'): str
                },
            }, 
            Optional('metric', default=None): {
                Optional('topk'): And(int, lambda s: s in [1, 5]),
                Optional('mAP'): {
                    Optional('anno_path'): str,
                    Optional('iou_thrs', default=0.5):
                            Or(And(str, lambda s: s in ['0.5:0.05:0.95']),
                               And(float, lambda s: s <= 1.0 and s >= 0.0)),
                    Optional('map_points', default=0): And(int, lambda s: s in [0, 11, 101])
                },
                Optional('COCOmAP'): {
                    Optional('anno_path'): str,
                    Optional('map_key', default='DetectionBoxes_Precision/mAP'): str
                },
                Optional('COCOmAPv2'): {
                    Optional('anno_path'): str,
                    Optional('map_key', default='DetectionBoxes_Precision/mAP'): str,
                    Optional('output_index_mapping', default={'num_detections': -1, 
                                                      'boxes': 0, 
                                                      'scores': 1, 
                                                      'classes': 2}): COCOmAP_input_order_schema
                },
                Optional('VOCmAP'): {
                    Optional('anno_path'): str
                },
                Optional('SquadF1'): Or({}, None),
                Optional('MSE'): {
                    Optional('compare_label'): bool
                },
                Optional('RMSE'): {
                    Optional('compare_label'): bool
                },
                Optional('MAE'): {
                    Optional('compare_label'): bool
                },
                Optional('Accuracy'): Or({}, None),
                Optional('Loss'): Or({}, None),
                Optional('BLEU'): Or({}, None),
                Optional('SquadF1'): Or({}, None),
                Optional('F1'): Or({}, None),
                Optional('mIOU'): {
                    Optional('num_classes'): int
                },
                Optional('GLUE'): {
                    Optional('task'): str
                },
                Optional('ROC'): {
                    Optional('task'): str
                },
            },
            Optional('configs'): configs_schema,
            Optional('iteration', default=-1): int,
            Optional('dataloader'): dataloader_schema,
            Optional('postprocess'): {
                Optional('transform'): postprocess_schema
            },
        },
        Optional('performance'): {
            Optional('warmup', default=5): int,
            Optional('iteration', default=-1): int,
            Optional('configs'): configs_schema,
            Optional('dataloader'): dataloader_schema,
            Optional('postprocess'): {
                Optional('transform'): postprocess_schema
            }
        },
    },
    Optional('pruning'): {
        Hook('train', handler=_valid_prune_epoch): object,
        Optional("train"): train_schema,
        Optional("approach"): approach_schema
    },

    Optional('distillation'): {
        Optional("train"): train_schema
    },

    Optional('auto_distillation'): {
        "search": {
            "search_space": dict,
            Optional("search_algorithm", default=""): str,
            Optional("metrics", default=[]): list,
            Optional("higher_is_better", default=[]): list,
            Optional("max_trials", default=1): int,
            Optional("seed", default=42): int,
            },
        Optional("flash_distillation"): {
            Optional("knowledge_transfer"): {
                Optional("block_names", default=[]): list,
                "layer_mappings_for_knowledge_transfer": list,
                Optional("loss_types", default=[]): list,
                Optional("loss_weights", default=[]): list,
                Optional("add_origin_loss", default=[]): list,
                Optional("train_steps", default=[]): list,
                },
            Optional("regular_distillation"): {
                Optional("block_names", default=[]): list,
                "layer_mappings_for_knowledge_transfer": list,
                Optional("loss_types", default=[]): list,
                Optional("loss_weights", default=[]): list,
                Optional("add_origin_loss", default=[]): list,
                Optional("train_steps", default=[]): list,
                },
            },
    },

    Optional('nas'): {
        Optional("approach", default=None): Or(str, None),
        Optional("search"): {
            Optional("search_space", default=None): Or(dict, None),
            Optional("search_algorithm", default=None): Or(str, None),
            Optional("metrics", default=None): list,
            Optional("higher_is_better", default=None): list,
            Optional("max_trials", default=None): int,
            Optional("seed", default=42): int,
            },
        Optional("dynas"): {
            Optional("supernet", default=None): str,
            Optional("metrics", default=None): list,
            Optional("population", default=50): int,
            Optional("num_evals", default=100000): int,
            Optional("results_csv_path", default=None): str,
            Optional("dataset_path", default=None): str,
            Optional("batch_size", default=64): int,
            },
    },

    Optional("train"): train_schema
})

quantization_default_schema = Schema({
    Optional('model', default={'name': 'default_model_name', \
                               'framework': 'NA', \
                                'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('quantization', default={'approach': 'post_training_static_quant', \
                                      'calibration': {'sampling_size': [100]},
                                      'recipes': {'scale_propagation_max_pooling': True,
                                                      'scale_propagation_concat': True,
                                                      'first_conv_or_matmul_quantization': True},
                                      'model_wise': {'weight': {'bit': [7.0]},
                                                     'activation': {}},
                                    }): dict,

    Optional('tuning', default={
        'strategy': {'name': 'basic'},
        'accuracy_criterion': {'relative': 0.01, 'higher_is_better': True},
        'objective': 'performance',
        'exit_policy': {'timeout': 0, 'max_trials': 100, 'performance_only': False},
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}): dict
})

pruning_default_schema = Schema({
    Optional('model', default={'name': 'default_model_name', \
                               'framework': 'NA', \
                                'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('tuning', default={
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('pruning', default={'approach': {'weight_compression':{'initial_sparsity': 0.0, \
                                            'target_sparsity': 0.97, 'start_epoch': 0, \
                                            'end_epoch': 4}}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}): dict
})

graph_optimization_default_schema = Schema({
    Optional('model', default={'name': 'resnet50', \
                               'framework': 'NA', \
                                'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('quantization', default={'approach': 'post_training_static_quant', 
                                    'calibration': {'sampling_size': [100]},
                                    'recipes': {'scale_propagation_max_pooling': True,
                                                    'scale_propagation_concat': True,
                                                    'first_conv_or_matmul_quantization': True},
                                    'model_wise': {'weight': {'bit': [7.0]},
                                                    'activation': {}}}): dict,

    Optional('tuning', default={
        'strategy': {'name': 'basic'},
        'accuracy_criterion': {'relative': 0.01, 'higher_is_better': True},
        'objective': 'performance',
        'exit_policy': {'timeout': 0, 'max_trials': 100, 'performance_only': False},
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}): dict,

    Optional('graph_optimization', default={'precisions': ['bf16, fp32']}): dict 
})

mixed_precision_default_schema = Schema({
    Optional('model', default={'name': 'resnet50', \
                               'framework': 'NA', \
                                'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('quantization', default={'approach': 'post_training_static_quant', 
                                    'calibration': {'sampling_size': [100]},
                                    'recipes': {'scale_propagation_max_pooling': True,
                                                    'scale_propagation_concat': True,
                                                    'first_conv_or_matmul_quantization': True},
                                    'model_wise': {'weight': {'bit': [7.0]},
                                                    'activation': {}}}): dict,

    Optional('tuning', default={
        'strategy': {'name': 'basic'},
        'accuracy_criterion': {'relative': 0.01, 'higher_is_better': True},
        'objective': 'performance',
        'exit_policy': {'timeout': 0, 'max_trials': 100, 'performance_only': False},
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}): dict,

    Optional('mixed_precision', default={'precisions': ['bf16, fp32']}): dict 
})

benchmark_default_schema = Schema({
    Optional('model', default={'name': 'resnet50', \
                               'framework': 'NA', \
                                'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('quantization', default={'approach': 'post_training_static_quant', 
                                    'calibration': {'sampling_size': [100]},
                                    'recipes': {'scale_propagation_max_pooling': True,
                                                    'scale_propagation_concat': True,
                                                    'first_conv_or_matmul_quantization': True},
                                    'model_wise': {'weight': {'bit': [7.0]},
                                                    'activation': {}}}): dict,

    Optional('tuning', default={
        'strategy': {'name': 'basic'},
        'accuracy_criterion': {'relative': 0.01, 'higher_is_better': True},
        'objective': 'performance',
        'exit_policy': {'timeout': 0, 'max_trials': 100, 'performance_only': False},
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}): dict
})

distillation_default_schema = Schema({
    Optional('model', default={'name': 'default_model_name', \
                               'framework': 'NA', \
                               'inputs': [], 'outputs': []}): dict,

    Optional('version', default=float(__version__.split('.')[0])): str,

    Optional('device', default='cpu'): str,

    Optional('tuning', default={
        'random_seed': 1978, 'tensorboard': False,
        'workspace': {'path': default_workspace}}): dict,

    Optional('distillation', default={
        'train': {'start_epoch': 0, 'end_epoch': 10, 
                  'iteration': 1000, 'frequency': 1, 
                  'optimizer': {'SGD': {'learning_rate': 0.001}}, 
                  'criterion': {'KnowledgeDistillationLoss': 
                                 {'temperature': 1.0, 
                                  'loss_types': ['CE', 'KL'], 
                                  'loss_weights': [0.5, 0.5]}}}}): dict,

    Optional('evaluation', default={'accuracy': {'metric': {'topk': 1}}}):dict
 
})

class Conf(object):
    """config parser.

    Args:
        cfg_fname (string): The path to the configuration file.

    """
    def __init__(self, cfg_fname):
        assert cfg_fname is not None
        self.usr_cfg = DotDict(self._read_cfg(cfg_fname))

    def _read_cfg(self, cfg_fname):
        """Load a config file following yaml syntax.

           Args:
               cfg_fname(string): The name of configuration yaml file
        """
        try:
            with open(cfg_fname, 'r') as f:
                content = f.read()
                cfg = yaml.safe_load(content)
                validated_cfg = schema.validate(cfg)

            # if user yaml doesn't include version field, neural_compressor will write a supported version
            # into it.
            if 'version' not in cfg:
                leading_whitespace = re.search(r"[ \t]*model\s*:",
                                               content).group().split("model")[0]
                content = re.sub(r'model\s*:',
                                 'version: {}\n\n{}model:'.format(
                                                               float(__version__.split('.')[0]),
                                                               leading_whitespace
                                                           ),
                                 content)
                with open(cfg_fname, 'w') as f:
                    f.write(content)

            return validated_cfg

        except Exception as e:
            logger.error("{}.".format(e))
            raise RuntimeError(
                "The yaml file format is not correct. Please refer to document."
            )

    def _convert_cfg(self, src, dst):
        """Helper function to merge user defined dict into default dict.

           If the key in src doesn't exist in dst, then add this key and value
           pair to dst.
           If the key in src is in dst, then override the value in dst with the
           value in src.

        Args:
            src (dict): The source dict merged from
            dst (dict): The source dict merged to

        Returns:
            dict: The merged dict from src to dst
        """
        for key in src:
            if key in dst:
                if isinstance(dst[key], dict) and isinstance(src[key], dict):
                    if key in ['accuracy_criterion', 'metric', 'dataset', 
                        'criterion', 'optimizer']:
                        # accuracy_criterion can only have one of absolute and relative
                        # others can only have one item
                        inter_key = src[key].keys() & dst[key].keys()-{'higher_is_better'}
                        if len(inter_key) == 0:
                            dst[key] = {}
                    if key == 'accuracy' and src[key].get('multi_metrics', None):
                        dst[key].pop('metric', None)
                    self._convert_cfg(src[key], dst[key])
                elif dst[key] == src[key]:
                    pass  # same leaf value
                else:
                    dst[key] = src[key]
            elif isinstance(src[key], dict):
                dst[key] = DotDict(self._convert_cfg(src[key], {}))
            else:
                dst[key] = src[key]
        return dst

class Quantization_Conf(Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(schema.validate(self._convert_cfg(
                cfg, copy.deepcopy(quantization_default_schema.validate(dict())))))
        else:
            self.usr_cfg = DotDict(quantization_default_schema.validate(dict()))
        self._model_wise_tune_space = None
        self._opwise_tune_space = None

    def _merge_dicts(self, src, dst):
        """Helper function to merge src dict into dst dict.

           If the key in src doesn't exist in dst, then skip
           If the key in src is in dst and the value intersects with the one in
           dst, then override the value in dst with the intersect value.

        Args:
            src (dict): The source dict merged from
            dst (dict): The source dict merged to

        Returns:
            dict: The merged dict from src to dst
        """
        for key in src:
            if key in dst:
                if isinstance(dst[key], dict) and isinstance(src[key], dict):
                    self._merge_dicts(src[key], dst[key])
                elif dst[key] == src[key] or src[key] is None:
                    pass  # same leaf value
                else:
                    value = [value for value in src[key]
                             if value in dst[key] or isinstance(value, float)]
                    if value != []:
                        dst[key] = value

        return dst

    def modelwise_tune_space(self, model_wise_quant):
        cfg = self.usr_cfg

        self._model_wise_tune_space = OrderedDict()
        for optype in model_wise_quant.keys():
            if cfg.quantization.optype_wise and optype in cfg.quantization.optype_wise:
                self._model_wise_tune_space[optype] = self._merge_dicts(
                    cfg.quantization.optype_wise[optype],
                    model_wise_quant[optype])
            else:
                self._model_wise_tune_space[optype] = self._merge_dicts(
                    cfg.quantization.model_wise,
                    model_wise_quant[optype])

        return self._model_wise_tune_space

    def _weight_compute(self, combined_cfg):
        temp_set = set()
        for _, config in combined_cfg.items():
            temp_str = ''
            for part, params in config.items():
                temp_str = temp_str + part
                for _, param in params.items():
                    temp_str += str(param)
                temp_str += '_'
            temp_set.add(temp_str)
        return len(temp_set)

    def _sort_cfgs(self, combined_cfgs):
        cfgs_num = len(combined_cfgs)
        for i in range(cfgs_num):
            for j in range(cfgs_num-i-1):
                weight_a = self._weight_compute(combined_cfgs[j])
                weight_b = self._weight_compute(combined_cfgs[j+1])
                if weight_a > weight_b:
                    temp = combined_cfgs[j]
                    combined_cfgs[j] = combined_cfgs[j+1]
                    combined_cfgs[j+1] = temp
        return combined_cfgs

    def _combine_optype_quant_cfgs(self, model_wise_quant_cfgs):
        if len(model_wise_quant_cfgs) == 0:
            return []
        temp_cfgs = OrderedDict()

        for optype, cfgs in model_wise_quant_cfgs.items():
            if len(cfgs) > 0:
                temp_cfgs[optype] = copy.deepcopy(cfgs)

        if not bool(temp_cfgs):
            return []

        keys, values = zip(*temp_cfgs.items())

        return self._sort_cfgs([dict(zip(keys, v)) for v in itertools.product(*values)])

    def opwise_tune_space(self, opwise_quant):
        def _is_regex(pattern):
            if re.match("^[A-Za-z0-9.][A-Za-z0-9_.\\-/]*$", pattern):
                return False
            return True

        opwise = copy.deepcopy(opwise_quant)

        cfg = self.usr_cfg
        if cfg.quantization.op_wise:
            for k, v in cfg.quantization.op_wise.items():
                is_regex = _is_regex(k)
                for k_op, _ in opwise.items():
                    if (not is_regex and k == k_op[0]) or (is_regex and re.match(k, k_op[0])):
                        opwise[k_op] = self._merge_dicts(v, opwise[k_op])

        for k, v in opwise.items():
            opwise[k] = self._merge_dicts(self._model_wise_tune_space[k[1]], opwise[k])

        self._opwise_tune_space = opwise
        return self._opwise_tune_space

    def expand_tune_cfgs(self, tune_space, framework):
        """generate all possible tuning combinations for each op or model wise tuning.

        Args:
            tune_space (dict): The tuning space to be expanded.

        Returns:
            dict: The expanded tuning configs
        """
        cfg_lists = self._expand_tune_cfgs_recursively(tune_space)

        # remove unreasonable tuning combinations
        valid_cfgs = []
        quant_dtype = ['int8', 'uint8', 'int4', 'uint4']

        for cfg in cfg_lists:
            cfg = DotDict(cfg)
            dtype = cfg.activation.dtype

            if dtype not in quant_dtype and "pytorch" not in framework:
                cfg.activation.clear()
                cfg.activation.dtype = dtype

            if 'weight' in cfg:
                dtype = cfg.weight.dtype
                if dtype not in quant_dtype:
                    cfg.weight.clear()
                    cfg.weight.dtype = dtype
                if (cfg.weight.dtype != cfg.activation.dtype and
                    cfg.weight.dtype not in quant_dtype and
                    cfg.activation.dtype not in quant_dtype) or \
                   (cfg.weight.dtype != cfg.activation.dtype and
                    cfg.weight.dtype in quant_dtype and
                    cfg.activation.dtype not in quant_dtype and
                    ("pytorch" not in framework or
                     len(tune_space['activation']['dtype']) > 1)) or \
                   (cfg.weight.dtype != cfg.activation.dtype and
                    cfg.weight.dtype not in quant_dtype and
                    cfg.activation.dtype in quant_dtype):
                    continue

            valid_cfgs.append(cfg)

        # remove duplicated configurations
        valid_cfgs = [cfg[0] for cfg in itertools.groupby(valid_cfgs)]
        return valid_cfgs

    def _expand_tune_cfgs_recursively(self, cfg_dict):
        """Helper function of recursively generating all combinations.

        Args:
            cfg_dict (dict): The dict of conf space.

        Returns:
            list: List containing all combinations
        """
        assert isinstance(cfg_dict, dict)
        combinations = OrderedDict()
        for key in cfg_dict:
            if isinstance(cfg_dict[key], dict):
                lists = self._expand_tune_cfgs_recursively(cfg_dict[key])
                combinations[key] = lists

        if len(combinations) != 0:
            return self._expand_tune_cfgs_recursively(combinations)

        keys, values = zip(*cfg_dict.items())
        values = list(filter(None, values))
        lists = [dict(zip(keys, v)) for v in itertools.product(*values)]
        return lists

class Pruning_Conf(Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(schema.validate(self._convert_cfg(
                cfg, copy.deepcopy(pruning_default_schema.validate(dict())))))
        else:
            self.usr_cfg = DotDict(pruning_default_schema.validate(dict()))

class Graph_Optimization_Conf(Quantization_Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(schema.validate(self._convert_cfg(
                cfg, copy.deepcopy(graph_optimization_default_schema.validate(dict())))))
        else:
            self.usr_cfg = DotDict(graph_optimization_default_schema.validate(dict()))

class MixedPrecision_Conf(Quantization_Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(self._convert_cfg(
                cfg, copy.deepcopy(mixed_precision_default_schema.validate(dict()))))
        else:
            self.usr_cfg = DotDict(mixed_precision_default_schema.validate(dict()))

class Benchmark_Conf(Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(schema.validate(self._convert_cfg(
                cfg, copy.deepcopy(benchmark_default_schema.validate(dict())))))
        else:
            self.usr_cfg = DotDict(benchmark_default_schema.validate(dict()))

class Distillation_Conf(Conf):
    """config parser.

    Args:
        cfg: The path to the configuration file or DotDict object or None.

    """

    def __init__(self, cfg=None):
        if isinstance(cfg, str):
            self.usr_cfg = DotDict(self._read_cfg(cfg))
        elif isinstance(cfg, DotDict):
            self.usr_cfg = DotDict(schema.validate(self._convert_cfg(
                cfg, copy.deepcopy(distillation_default_schema.validate(dict())))))
        else:
            self.usr_cfg = DotDict(distillation_default_schema.validate(dict()))

class NASConfig(Conf):
    """config parser.

    Args:
        approach: The approach of the NAS.
        search_algorithm: The search algorithm for NAS procedure.

    """

    def __init__(self, approach=None, search_space=None, search_algorithm=None):
        assert approach is None or (isinstance(approach, str) and approach), \
            "Should set approach with a string."
        usr_cfg = {
            'model': {'name': 'nas', 'framework': 'NA'},
            'nas': {'approach': approach,
                    'search': {
                        'search_space': search_space,
                        'search_algorithm': search_algorithm}
                    },
        }
        self.usr_cfg = DotDict(usr_cfg)
        if approach and approach != 'basic':
            self.usr_cfg.nas[approach] = DotDict({})
            self.__setattr__(approach, self.usr_cfg.nas[approach])

    def validate(self):
        self.usr_cfg = schema.validate(self.usr_cfg)
        
    @property
    def nas(self):
        return self.usr_cfg.nas

class DefaultConf(DotDict):
    def __getitem__(self, key):
        if key not in self:
            self[key] = DefaultConf({})
        value = self.get(key, None)
        return value

    __getattr__ = __getitem__

conf = DefaultConf({})

QuantConf = Quantization_Conf
PruningConf = Pruning_Conf
GraphOptConf = Graph_Optimization_Conf
BenchmarkConf = Benchmark_Conf
DistillationConf = Distillation_Conf
MixedPrecisionConf = MixedPrecision_Conf
